import asyncio
import threading
import time
from datetime import datetime
from decimal import Decimal
from typing import TYPE_CHECKING, List, Optional, Set, Tuple

import pandas as pd

from hummingbot.client.performance import PerformanceMetrics
from hummingbot.client.settings import MAXIMUM_TRADE_FILLS_DISPLAY_OUTPUT, AllConnectorSettings
from hummingbot.client.ui.interface_utils import format_df_for_printout
from hummingbot.core.utils.async_utils import safe_ensure_future
from hummingbot.model.trade_fill import TradeFill

s_float_0 = float(0)
s_decimal_0 = Decimal("0")


if TYPE_CHECKING:
    from hummingbot.client.hummingbot_application import HummingbotApplication  # noqa: F401


def get_timestamp(days_ago: float = 0.) -> float:
    return time.time() - (60. * 60. * 24. * days_ago)


class HistoryCommand:
    def history(self,  # type: HummingbotApplication
                days: float = 0,
                verbose: bool = False,
                precision: Optional[int] = None
                ):
        if threading.current_thread() != threading.main_thread():
            self.ev_loop.call_soon_threadsafe(self.history, days, verbose, precision)
            return

        if self.strategy_file_name is None:
            self.notify("\n  Please first import a strategy config file of which to show historical performance.")
            return
        start_time = get_timestamp(days) if days > 0 else self.init_time
        with self.trading_core.trade_fill_db.get_new_session() as session:
            trades: List[TradeFill] = self._get_trades_from_session(
                int(start_time * 1e3),
                session=session,
                config_file_path=self.strategy_file_name)
            if not trades:
                self.notify("\n  No past trades to report.")
                return
            if verbose:
                self.list_trades(start_time)
            safe_ensure_future(self.history_report(start_time, trades, precision))

    def get_history_trades_json(self,  # type: HummingbotApplication
                                days: float = 0):
        if self.strategy_file_name is None:
            return
        start_time = get_timestamp(days) if days > 0 else self.init_time
        with self.trading_core.trade_fill_db.get_new_session() as session:
            trades: List[TradeFill] = self._get_trades_from_session(
                int(start_time * 1e3),
                session=session,
                config_file_path=self.strategy_file_name)
            return list([TradeFill.to_bounty_api_json(t) for t in trades])

    async def history_report(self,  # type: HummingbotApplication
                             start_time: float,
                             trades: List[TradeFill],
                             precision: Optional[int] = None,
                             display_report: bool = True) -> Decimal:
        market_info: Set[Tuple[str, str]] = set((t.market, t.symbol) for t in trades)
        if display_report:
            self.report_header(start_time)
        return_pcts = []
        for market, symbol in market_info:
            cur_trades = [t for t in trades if t.market == market and t.symbol == symbol]
            network_timeout = float(self.client_config_map.commands_timeout.other_commands_timeout)
            try:
                cur_balances = await asyncio.wait_for(self.trading_core.get_current_balances(market), network_timeout)
            except asyncio.TimeoutError:
                self.notify(
                    "\nA network error prevented the balances retrieval to complete. See logs for more details."
                )
                raise
            perf = await PerformanceMetrics.create(symbol, cur_trades, cur_balances)
            if display_report:
                self.report_performance_by_market(market, symbol, perf, precision)
            return_pcts.append(perf.return_pct)
        avg_return = sum(return_pcts) / len(return_pcts) if len(return_pcts) > 0 else s_decimal_0
        if display_report and len(return_pcts) > 1:
            self.notify(f"\nAveraged Return = {avg_return:.2%}")
        return avg_return

    def report_header(self,  # type: HummingbotApplication
                      start_time: float):
        lines = []
        current_time = get_timestamp()
        lines.extend(
            [f"\nStart Time: {datetime.fromtimestamp(start_time).strftime('%Y-%m-%d %H:%M:%S')}"] +
            [f"Current Time: {datetime.fromtimestamp(current_time).strftime('%Y-%m-%d %H:%M:%S')}"] +
            [f"Duration: {pd.Timedelta(seconds=int(current_time - start_time))}"]
        )
        self.notify("\n".join(lines))

    def report_performance_by_market(self,  # type: HummingbotApplication
                                     market: str,
                                     trading_pair: str,
                                     perf: PerformanceMetrics,
                                     precision: int):
        lines = []
        base, quote = trading_pair.split("-")
        lines.extend(
            [f"\n{market} / {trading_pair}"]
        )

        trades_columns = ["", "buy", "sell", "total"]
        trades_data = [
            [f"{'Number of trades':<27}", perf.num_buys, perf.num_sells, perf.num_trades],
            [f"{f'Total trade volume ({base})':<27}",
             PerformanceMetrics.smart_round(perf.b_vol_base, precision),
             PerformanceMetrics.smart_round(perf.s_vol_base, precision),
             PerformanceMetrics.smart_round(perf.tot_vol_base, precision)],
            [f"{f'Total trade volume ({quote})':<27}",
             PerformanceMetrics.smart_round(perf.b_vol_quote, precision),
             PerformanceMetrics.smart_round(perf.s_vol_quote, precision),
             PerformanceMetrics.smart_round(perf.tot_vol_quote, precision)],
            [f"{'Avg price':<27}",
             PerformanceMetrics.smart_round(perf.avg_b_price, precision),
             PerformanceMetrics.smart_round(perf.avg_s_price, precision),
             PerformanceMetrics.smart_round(perf.avg_tot_price, precision)],
        ]
        trades_df: pd.DataFrame = pd.DataFrame(data=trades_data, columns=trades_columns)
        lines.extend(["", "  Trades:"] + ["    " + line for line in trades_df.to_string(index=False).split("\n")])

        assets_columns = ["", "start", "current", "change"]
        assets_data = [
            [f"{base:<17}", "-", "-", "-"] if market in AllConnectorSettings.get_derivative_names() else  # No base asset for derivatives because they are margined
            [f"{base:<17}",
             PerformanceMetrics.smart_round(perf.start_base_bal, precision),
             PerformanceMetrics.smart_round(perf.cur_base_bal, precision),
             PerformanceMetrics.smart_round(perf.tot_vol_base, precision)],
            [f"{quote:<17}",
             PerformanceMetrics.smart_round(perf.start_quote_bal, precision),
             PerformanceMetrics.smart_round(perf.cur_quote_bal, precision),
             PerformanceMetrics.smart_round(perf.tot_vol_quote, precision)],
            [f"{trading_pair + ' price':<17}",
             PerformanceMetrics.smart_round(perf.start_price),
             PerformanceMetrics.smart_round(perf.cur_price),
             PerformanceMetrics.smart_round(perf.cur_price - perf.start_price)],
            [f"{'Base asset %':<17}", "-", "-", "-"] if market in AllConnectorSettings.get_derivative_names() else  # No base asset for derivatives because they are margined
            [f"{'Base asset %':<17}",
             f"{perf.start_base_ratio_pct:.2%}",
             f"{perf.cur_base_ratio_pct:.2%}",
             f"{perf.cur_base_ratio_pct - perf.start_base_ratio_pct:.2%}"],
        ]
        assets_df: pd.DataFrame = pd.DataFrame(data=assets_data, columns=assets_columns)
        lines.extend(["", "  Assets:"] + ["    " + line for line in assets_df.to_string(index=False).split("\n")])

        perf_data = [
            ["Hold portfolio value    ", f"{PerformanceMetrics.smart_round(perf.hold_value, precision)} {quote}"],
            ["Current portfolio value ", f"{PerformanceMetrics.smart_round(perf.cur_value, precision)} {quote}"],
            ["Trade P&L               ", f"{PerformanceMetrics.smart_round(perf.trade_pnl, precision)} {quote}"]
        ]
        perf_data.extend(
            ["Fees paid               ", f"{PerformanceMetrics.smart_round(fee_amount, precision)} {fee_token}"]
            for fee_token, fee_amount in perf.fees.items()
        )
        perf_data.extend(
            [["Total P&L               ", f"{PerformanceMetrics.smart_round(perf.total_pnl, precision)} {quote}"],
             ["Return %                ", f"{perf.return_pct:.2%}"]]
        )
        perf_df: pd.DataFrame = pd.DataFrame(data=perf_data)
        lines.extend(["", "  Performance:"] +
                     ["    " + line for line in perf_df.to_string(index=False, header=False).split("\n")])

        self.notify("\n".join(lines))

    def list_trades(self,  # type: HummingbotApplication
                    start_time: float):
        if threading.current_thread() != threading.main_thread():
            self.ev_loop.call_soon_threadsafe(self.list_trades, start_time)
            return

        lines = []

        with self.trading_core.trade_fill_db.get_new_session() as session:
            queried_trades: List[TradeFill] = self._get_trades_from_session(
                int(start_time * 1e3),
                session=session,
                number_of_rows=MAXIMUM_TRADE_FILLS_DISPLAY_OUTPUT + 1,
                config_file_path=self.strategy_file_name)
            df: pd.DataFrame = TradeFill.to_pandas(queried_trades)

        if len(df) > 0:
            # Check if number of trades exceed maximum number of trades to display
            if len(df) > MAXIMUM_TRADE_FILLS_DISPLAY_OUTPUT:
                df = df[:MAXIMUM_TRADE_FILLS_DISPLAY_OUTPUT]
                self.notify(
                    f"\n  Showing last {MAXIMUM_TRADE_FILLS_DISPLAY_OUTPUT} trades in the current session.")
            df_lines = format_df_for_printout(df, self.client_config_map.tables_format).split("\n")
            lines.extend(["", "  Recent trades:"] +
                         ["    " + line for line in df_lines])
        else:
            lines.extend(["\n  No past trades in this session."])
        self.notify("\n".join(lines))



# A single source of truth for constant variables related to the exchange
from hummingbot.core.api_throttler.data_types import LinkedLimitWeightPair, RateLimit

EXCHANGE_NAME = "gate_io"
DEFAULT_DOMAIN = ""
HBOT_BROKER_ID = "hummingbot"
HBOT_ORDER_ID = "t-HBOT"
MAX_ID_LEN = 30

REST_URL = "https://api.gateio.ws/api/v4"
REST_URL_AUTH = "/api/v4"
WS_URL = "wss://api.gateio.ws/ws/v4/"
NETWORK_CHECK_PATH_URL = "spot/currencies/BTC"
SYMBOL_PATH_URL = "spot/currency_pairs"
ORDER_CREATE_PATH_URL = "spot/orders"
ORDER_DELETE_PATH_URL = "spot/orders/{order_id}"
USER_BALANCES_PATH_URL = "spot/accounts"
ORDER_STATUS_PATH_URL = "spot/orders/{order_id}"
USER_ORDERS_PATH_URL = "spot/open_orders"
TICKER_PATH_URL = "spot/tickers"
ORDER_BOOK_PATH_URL = "spot/order_book"
MY_TRADES_PATH_URL = "spot/my_trades"
SERVER_TIME_URL = "spot/time"

TRADES_ENDPOINT_NAME = "spot.trades"
ORDER_SNAPSHOT_ENDPOINT_NAME = "spot.order_book"
ORDERS_UPDATE_ENDPOINT_NAME = "spot.order_book_update"
USER_TRADES_ENDPOINT_NAME = "spot.usertrades"
USER_ORDERS_ENDPOINT_NAME = "spot.orders"
USER_BALANCE_ENDPOINT_NAME = "spot.balances"
PONG_CHANNEL_NAME = "spot.pong"

# Timeouts
MESSAGE_TIMEOUT = 30.0
PING_TIMEOUT = 10.0
API_CALL_TIMEOUT = 10.0
API_MAX_RETRIES = 4

# Intervals
# Only used when nothing is received from WS
SHORT_POLL_INTERVAL = 5.0
# 45 seconds should be fine since we get trades, orders and balances via WS
LONG_POLL_INTERVAL = 45.0
# One minute should be fine since we get trades, orders and balances via WS
UPDATE_ORDER_STATUS_INTERVAL = 60.0
# 10 minute interval to update trading rules, these would likely never change whilst running.
INTERVAL_TRADING_RULES = 600

PUBLIC_URL_POINTS_LIMIT_ID = "PublicPoints"
PRIVATE_URL_POINTS_LIMIT_ID = "PrivatePoints"  # includes place-orders
CANCEL_ORDERS_LIMITS_ID = "CancelOrders"
ORDER_DELETE_LIMIT_ID = "OrderDelete"
ORDER_STATUS_LIMIT_ID = "OrderStatus"
RATE_LIMITS = [
    RateLimit(limit_id=PUBLIC_URL_POINTS_LIMIT_ID, limit=900, time_interval=1),
    RateLimit(limit_id=PRIVATE_URL_POINTS_LIMIT_ID, limit=900, time_interval=1),
    RateLimit(limit_id=CANCEL_ORDERS_LIMITS_ID, limit=5_000, time_interval=1),
    RateLimit(limit_id=NETWORK_CHECK_PATH_URL, limit=900, time_interval=1, linked_limits=[LinkedLimitWeightPair(PUBLIC_URL_POINTS_LIMIT_ID)]),
    RateLimit(limit_id=SYMBOL_PATH_URL, limit=900, time_interval=1, linked_limits=[LinkedLimitWeightPair(PUBLIC_URL_POINTS_LIMIT_ID)]),
    RateLimit(limit_id=ORDER_CREATE_PATH_URL, limit=900, time_interval=1, linked_limits=[LinkedLimitWeightPair(PRIVATE_URL_POINTS_LIMIT_ID)]),
    RateLimit(limit_id=ORDER_DELETE_LIMIT_ID, limit=5_000, time_interval=1, linked_limits=[LinkedLimitWeightPair(CANCEL_ORDERS_LIMITS_ID)]),
    RateLimit(limit_id=USER_BALANCES_PATH_URL, limit=900, time_interval=1, linked_limits=[LinkedLimitWeightPair(PRIVATE_URL_POINTS_LIMIT_ID)]),
    RateLimit(limit_id=ORDER_STATUS_LIMIT_ID, limit=900, time_interval=1, linked_limits=[LinkedLimitWeightPair(PRIVATE_URL_POINTS_LIMIT_ID)]),
    RateLimit(limit_id=USER_ORDERS_PATH_URL, limit=900, time_interval=1, linked_limits=[LinkedLimitWeightPair(PRIVATE_URL_POINTS_LIMIT_ID)]),
    RateLimit(limit_id=TICKER_PATH_URL, limit=900, time_interval=1, linked_limits=[LinkedLimitWeightPair(PUBLIC_URL_POINTS_LIMIT_ID)]),
    RateLimit(limit_id=ORDER_BOOK_PATH_URL, limit=900, time_interval=1, linked_limits=[LinkedLimitWeightPair(PUBLIC_URL_POINTS_LIMIT_ID)]),
    RateLimit(limit_id=MY_TRADES_PATH_URL, limit=900, time_interval=1, linked_limits=[LinkedLimitWeightPair(PRIVATE_URL_POINTS_LIMIT_ID)]),
    RateLimit(limit_id=SERVER_TIME_URL, limit=900, time_interval=1, linked_limits=[LinkedLimitWeightPair(PUBLIC_URL_POINTS_LIMIT_ID)]),
]

# ERROR LABELS, see https://www.gate.io/docs/developers/apiv4/#label-list
ERR_LABEL_ORDER_NOT_FOUND = "ORDER_NOT_FOUND"
ERR_LABEL_TIME_RELATED_ERROR = "REQUEST_EXPIRED"



import json
import random
import re
from abc import ABC, abstractmethod
from decimal import Decimal
from pathlib import Path
from typing import TYPE_CHECKING, Any, Callable, Dict, List, Union

from pydantic import ConfigDict, Field, SecretStr, field_validator, model_validator
from tabulate import tabulate_formats

from hummingbot.client.config.config_data_types import BaseClientModel, ClientConfigEnum
from hummingbot.client.config.config_methods import using_exchange as using_exchange_pointer
from hummingbot.client.config.config_validators import validate_bool, validate_float
from hummingbot.client.settings import DEFAULT_LOG_FILE_PATH, AllConnectorSettings
from hummingbot.connector.connector_base import ConnectorBase
from hummingbot.connector.connector_metrics_collector import (
    DummyMetricsCollector,
    MetricsCollector,
    TradeVolumeMetricCollector,
)
from hummingbot.connector.exchange.binance.binance_utils import BinanceConfigMap
from hummingbot.connector.exchange.gate_io.gate_io_utils import GateIOConfigMap
from hummingbot.connector.exchange.kraken.kraken_utils import KrakenConfigMap
from hummingbot.connector.exchange.kucoin.kucoin_utils import KuCoinConfigMap
from hummingbot.core.rate_oracle.rate_oracle import RATE_ORACLE_SOURCES, RateOracle
from hummingbot.core.rate_oracle.sources.rate_source_base import RateSourceBase
from hummingbot.core.utils.kill_switch import ActiveKillSwitch, KillSwitch, PassThroughKillSwitch

if TYPE_CHECKING:
    from hummingbot.core.trading_core import TradingCore


def generate_client_id() -> str:
    vals = [random.choice(range(0, 256)) for i in range(0, 20)]
    return "".join([f"{val:02x}" for val in vals])


def using_exchange(exchange: str) -> Callable:
    return using_exchange_pointer(exchange)


class MQTTBridgeConfigMap(BaseClientModel):
    mqtt_host: str = Field(
        default="localhost",
        json_schema_extra={"prompt": lambda cm: "Set the MQTT hostname to connect to (e.g. localhost)"}
    )
    mqtt_port: int = Field(
        default=1883,
        json_schema_extra={"prompt": lambda cm: "Set the MQTT port to connect to (e.g. 1883)"},
    )
    mqtt_username: str = Field(
        default="",
        json_schema_extra={"prompt": lambda cm: "Set the username for connecting to the MQTT broker"},
    )
    mqtt_password: str = Field(
        default="",
        json_schema_extra={"prompt": lambda cm: "Set the password for connecting to the MQTT broker"},
    )
    mqtt_namespace: str = Field(
        default='hbot',
        json_schema_extra={"prompt": lambda cm: "Set the MQTT namespace to connect to (e.g. hbot)"},
    )
    mqtt_ssl: bool = Field(
        default=False,
        json_schema_extra={"prompt": lambda cm: "Enable/Disable SSL for MQTT connections"},
    )
    mqtt_logger: bool = Field(default=True)
    mqtt_notifier: bool = Field(
        default=True,
        json_schema_extra={"prompt": lambda cm: "Enable/Disable MQTT Notifier"},
    )
    mqtt_commands: bool = Field(
        default=True,
        json_schema_extra={"prompt": lambda cm: "Enable/Disable MQTT Commands"},
    )
    mqtt_events: bool = Field(
        default=True,
        json_schema_extra={"prompt": lambda cm: "Enable/Disable MQTT Events"},
    )
    mqtt_external_events: bool = Field(
        default=True,
        json_schema_extra={"prompt": lambda cm: "Enable/Disable External MQTT Events"},
    )
    mqtt_autostart: bool = Field(
        default=False,
        json_schema_extra={"prompt": lambda cm: "Enable/Disable MQTT Autostart"},
    )
    model_config = ConfigDict(title="mqtt_bridge")


class MarketDataCollectionConfigMap(BaseClientModel):
    market_data_collection_enabled: bool = Field(
        default=False,
        json_schema_extra={"prompt": lambda cm: "Enable/Disable Market Data Collection"},
    )
    market_data_collection_interval: int = Field(
        default=60,
        ge=1,
        json_schema_extra={"prompt": lambda cm: "Set the market data collection interval in seconds (Default=60)"},
    )
    market_data_collection_depth: int = Field(
        default=20,
        ge=2,
        json_schema_extra={"prompt": lambda cm: "Set the order book collection depth (Default=20)"},
    )
    model_config = ConfigDict(title="market_data_collection")


class ColorConfigMap(BaseClientModel):
    top_pane: str = Field(
        default="#000000",
        json_schema_extra={"prompt": lambda cm: "What is the background color of the top pane?"},
    )
    bottom_pane: str = Field(
        default="#000000",
        json_schema_extra={"prompt": lambda cm: "What is the background color of the bottom pane?"},
    )
    output_pane: str = Field(
        default="#262626",
        json_schema_extra={"prompt": lambda cm: "What is the background color of the output pane?"},
    )
    input_pane: str = Field(
        default="#1C1C1C",
        json_schema_extra={"prompt": lambda cm: "What is the background color of the input pane?"},
    )
    logs_pane: str = Field(
        default="#121212",
        json_schema_extra={"prompt": lambda cm: "What is the background color of the logs pane?"},
    )
    terminal_primary: str = Field(
        default="#5FFFD7",
        json_schema_extra={"prompt": lambda cm: "What is the terminal primary color?"},
    )
    primary_label: str = Field(
        default="#5FFFD7",
        json_schema_extra={"prompt": lambda cm: "What is the background color for primary label?"},
    )
    secondary_label: str = Field(
        default="#FFFFFF",
        json_schema_extra={"prompt": lambda cm: "What is the background color for secondary label?"},
    )
    success_label: str = Field(
        default="#5FFFD7",
        json_schema_extra={"prompt": lambda cm: "What is the background color for success label?"},
    )
    warning_label: str = Field(
        default="#FFFF00",
        json_schema_extra={"prompt": lambda cm: "What is the background color for warning label?"},
    )
    info_label: str = Field(
        default="#5FD7FF",
        json_schema_extra={"prompt": lambda cm: "What is the background color for info label?"},
    )
    error_label: str = Field(
        default="#FF0000",
        json_schema_extra={"prompt": lambda cm: "What is the background color for error label?"},
    )
    gold_label: str = Field(
        default="#FFD700",
        json_schema_extra={"prompt": lambda cm: "What is the background color for gold label?"},
    )
    silver_label: str = Field(
        default="#C0C0C0",
        json_schema_extra={"prompt": lambda cm: "What is the background color for silver label?"},
    )
    bronze_label: str = Field(
        default="#CD7F32",
        json_schema_extra={"prompt": lambda cm: "What is the background color for bronze label?"},
    )

    @field_validator(
        "top_pane",
        "bottom_pane",
        "output_pane",
        "input_pane",
        "logs_pane",
        "terminal_primary",
        "primary_label",
        "secondary_label",
        "success_label",
        "warning_label",
        "info_label",
        "error_label",
        "gold_label",
        "silver_label",
        "bronze_label",
        mode="before")
    @classmethod
    def validate_color(cls, v: str):
        if not re.search(r'^#(?:[0-9a-fA-F]{2}){3}$', v):
            raise ValueError("Invalid color code")
        return v


class PaperTradeConfigMap(BaseClientModel):
    paper_trade_exchanges: List = Field(
        default=[
            BinanceConfigMap.model_config["title"],
            KuCoinConfigMap.model_config["title"],
            KrakenConfigMap.model_config["title"],
            GateIOConfigMap.model_config["title"],
        ],
    )
    paper_trade_account_balance: Dict[str, float] = Field(
        default={
            "BTC": 1,
            "USDT": 100000,
            "USDC": 100000,
            "ETH": 20,
            "WETH": 20,
            "SOL": 100,
            "DOGE": 1000000,
            "HBOT": 10000000,
        },
        json_schema_extra={"prompt": lambda cm: (
            "Enter paper trade balance settings (Input must be valid json — "
            "e.g. {\"ETH\": 10, \"USDC\": 50000})"
        )},
    )

    @field_validator("paper_trade_account_balance", mode="before")
    @classmethod
    def validate_paper_trade_account_balance(cls, v: Union[str, Dict[str, float]]):
        if isinstance(v, str):
            v = json.loads(v)
        return v


class KillSwitchMode(BaseClientModel, ABC):
    @abstractmethod
    def get_kill_switch(self, trading_core: "TradingCore") -> KillSwitch:
        ...


class KillSwitchEnabledMode(KillSwitchMode):
    kill_switch_rate: Decimal = Field(
        default=Decimal("10"),
        json_schema_extra={
            "prompt": lambda cm: "At what profit/loss rate would you like the bot to stop? "
                                 "(e.g. -5 equals 5 percent loss)"
        }
    )
    model_config = ConfigDict(title="kill_switch_enabled")

    def get_kill_switch(self, trading_core: "TradingCore") -> ActiveKillSwitch:
        kill_switch = ActiveKillSwitch(kill_switch_rate=self.kill_switch_rate, trading_core=trading_core)
        return kill_switch


class KillSwitchDisabledMode(KillSwitchMode):
    model_config = ConfigDict(title="kill_switch_disabled")

    def get_kill_switch(self, trading_core: "TradingCore") -> PassThroughKillSwitch:
        kill_switch = PassThroughKillSwitch()
        return kill_switch


KILL_SWITCH_MODES = {
    KillSwitchEnabledMode.model_config["title"]: KillSwitchEnabledMode,
    KillSwitchDisabledMode.model_config["title"]: KillSwitchDisabledMode,
}


class AutofillImportEnum(str, ClientConfigEnum):
    start = "start"
    config = "config"
    disabled = "disabled"


class DBMode(BaseClientModel, ABC):
    @abstractmethod
    def get_url(self, db_path: str) -> str:
        ...


class DBSqliteMode(DBMode):
    db_engine: str = Field(
        default="sqlite",
        json_schema_extra={
            "prompt": lambda cm: "Please enter database engine you want to use "
                                 "(reference: https://docs.sqlalchemy.org/en/13/dialects/)"
        }
    )
    model_config = ConfigDict(title="sqlite_db_engine")

    def get_url(self, db_path: str) -> str:
        return f"{self.db_engine}:///{db_path}"


class DBOtherMode(DBMode):
    db_engine: str = Field(
        default=...,
        json_schema_extra={
            "prompt": lambda cm: "Please enter database engine you want to use "
        }
    )
    db_host: str = Field(
        default="127.0.0.1",
        json_schema_extra={"prompt": lambda cm: "Please enter your DB host address"},
    )
    db_port: int = Field(
        default=3306,
        json_schema_extra={"prompt": lambda cm: "Please enter your DB port"},
    )
    db_username: str = Field(
        default="username",
        json_schema_extra={"prompt": lambda cm: "Please enter your DB username"},
    )
    db_password: str = Field(
        default="password",
        json_schema_extra={"prompt": lambda cm: "Please enter your DB password"},
    )
    db_name: str = Field(
        default="dbname",
        json_schema_extra={"prompt": lambda cm: "Please enter your DB name"},
    )
    model_config = ConfigDict(title="other_db_engine")

    def get_url(self, db_path: str) -> str:
        return f"{self.db_engine}://{self.db_username}:{self.db_password}@{self.db_host}:{self.db_port}/{self.db_name}"

    @field_validator("db_engine")
    @classmethod
    def validate_db_engine(cls, v: str):
        assert v != "sqlite"
        return v


DB_MODES = {
    DBSqliteMode.model_config["title"]: DBSqliteMode,
    DBOtherMode.model_config["title"]: DBOtherMode,
}


class GatewayConfigMap(BaseClientModel):
    gateway_api_host: str = Field(
        default="localhost",
        json_schema_extra={"prompt": lambda cm: "Please enter your Gateway API host"},
    )
    gateway_api_port: str = Field(
        default="15888",
        json_schema_extra={"prompt": lambda cm: "Please enter your Gateway API port"},
    )
    gateway_use_ssl: bool = Field(
        default=False,
        json_schema_extra={"prompt": lambda cm: "Enable SSL endpoints for secure Gateway connection? (True / False)"},
    )

    model_config = ConfigDict(title="gateway")


class GlobalTokenConfigMap(BaseClientModel):
    global_token_name: str = Field(
        default="USDT",
        json_schema_extra={"prompt": lambda cm: "What is your default display token? (e.g. USDT, BTC)"},
    )
    global_token_symbol: str = Field(
        default="$",
        json_schema_extra={"prompt": lambda cm: "What is your default display token symbol? (e.g. $,€)"},
    )
    model_config = ConfigDict(title="global_token")

    @field_validator("global_token_name")
    @classmethod
    def validate_global_token_name(cls, v: str) -> str:
        return v.upper()

    # === post-validations ===

    @model_validator(mode="after")
    def post_validations(self):
        RateOracle.get_instance().quote_token = self.global_token_name
        return self


class CommandsTimeoutConfigMap(BaseClientModel):
    create_command_timeout: Decimal = Field(
        default=Decimal("10"),
        gt=Decimal("0"),
        json_schema_extra={
            "prompt": lambda cm: "Network timeout when fetching the minimum order amount in the create command (in seconds)"
        }
    )
    other_commands_timeout: Decimal = Field(
        default=Decimal("30"),
        gt=Decimal("0"),
        json_schema_extra={
            "prompt": lambda cm: "Network timeout to apply to the other commands' API calls (in seconds)"
        }
    )
    model_config = ConfigDict(title="commands_timeout")


class AnonymizedMetricsMode(BaseClientModel, ABC):
    @abstractmethod
    def get_collector(
            self,
            connector: ConnectorBase,
            rate_provider: RateOracle,
            instance_id: str,
            valuation_token: str = "USDT",
    ) -> MetricsCollector:
        ...


class AnonymizedMetricsDisabledMode(AnonymizedMetricsMode):
    model_config = ConfigDict(title="anonymized_metrics_disabled")

    def get_collector(
            self,
            connector: ConnectorBase,
            rate_provider: RateOracle,
            instance_id: str,
            valuation_token: str = "USDT",
    ) -> MetricsCollector:
        return DummyMetricsCollector()


class AnonymizedMetricsEnabledMode(AnonymizedMetricsMode):
    anonymized_metrics_interval_min: Decimal = Field(
        default=Decimal("15"),
        gt=Decimal("0"),
        json_schema_extra={"prompt": lambda cm: "How often do you want to send the anonymized metrics (in minutes)"},
    )
    model_config = ConfigDict(title="anonymized_metrics_enabled")

    def get_collector(
            self,
            connector: ConnectorBase,
            rate_provider: RateOracle,
            instance_id: str,
            valuation_token: str = "USDT",
    ) -> MetricsCollector:
        instance = TradeVolumeMetricCollector(
            connector=connector,
            activation_interval=self.anonymized_metrics_interval_min,
            rate_provider=rate_provider,
            instance_id=instance_id,
            valuation_token=valuation_token,
        )
        return instance


METRICS_MODES = {
    AnonymizedMetricsDisabledMode.model_config["title"]: AnonymizedMetricsDisabledMode,
    AnonymizedMetricsEnabledMode.model_config["title"]: AnonymizedMetricsEnabledMode,
}


class RateSourceModeBase(BaseClientModel, ABC):
    @abstractmethod
    def build_rate_source(self) -> RateSourceBase:
        ...


class ExchangeRateSourceModeBase(RateSourceModeBase):
    def build_rate_source(self) -> RateSourceBase:
        return RATE_ORACLE_SOURCES[self.model_config["title"]]()


class AscendExRateSourceMode(ExchangeRateSourceModeBase):
    name: str = Field(default="ascend_ex")
    model_config = ConfigDict(title="ascend_ex")


class BinanceRateSourceMode(ExchangeRateSourceModeBase):
    name: str = Field(default="binance")
    model_config = ConfigDict(title="binance")


class MexcRateSourceMode(ExchangeRateSourceModeBase):
    name: str = Field(default="mexc")
    model_config = ConfigDict(title="mexc")


class CubeRateSourceMode(ExchangeRateSourceModeBase):
    name: str = Field(default="cube")
    model_config = ConfigDict(title="cube")


class CoinGeckoRateSourceMode(RateSourceModeBase):
    name: str = Field(default="coin_gecko")
    extra_tokens: List[str] = Field(
        default=[],
        json_schema_extra={
            "prompt": lambda cm: (
                "List of comma-delimited CoinGecko token ids to always include"
                " in CoinGecko rates query (e.g. frontier-token,pax-gold,rbtc — empty to skip)"
            ),
        }
    )
    api_key: str = Field(
        default="",
        description="API key to use to request information from CoinGecko (if empty public API will be used)",
        json_schema_extra={
            "prompt": lambda cm: "CoinGecko API key (optional, leave empty to use public API) NOTE: will be stored in plain text due to a bug in the way hummingbot loads the config file",
            "prompt_on_new": True,
            "is_connect_key": True,
        },
    )

    api_tier: str = Field(
        default="PUBLIC",
        description="API tier for CoinGecko (PUBLIC, DEMO, or PRO)",
        json_schema_extra={
            "prompt": lambda cm: "Select CoinGecko API tier (PUBLIC/DEMO/PRO)",
            "prompt_on_new": True,
            "is_connect_key": True,
        },
    )
    model_config = ConfigDict(title="coin_gecko")

    def build_rate_source(self) -> RateSourceBase:
        return self._build_rate_source_cls(
            extra_tokens=self.extra_tokens,
            api_key=self.api_key,
            api_tier=self.api_tier
        )

    @field_validator("extra_tokens", mode="before")
    def validate_extra_tokens(cls, value: Union[str, List[str]]):
        extra_tokens = value.split(",") if isinstance(value, str) else value
        return extra_tokens

    @field_validator("api_tier", mode="before")
    def validate_api_tier(cls, v: str):
        from hummingbot.data_feed.coin_gecko_data_feed.coin_gecko_constants import CoinGeckoAPITier
        valid_tiers = [tier.name for tier in CoinGeckoAPITier]
        if v.upper() not in valid_tiers:
            return CoinGeckoAPITier.PUBLIC.name
        return v.upper()

    @model_validator(mode="after")
    def post_validations(self):
        RateOracle.get_instance().source = self.build_rate_source()
        return self

    @classmethod
    def _build_rate_source_cls(cls, extra_tokens: List[str], api_key: str, api_tier: str) -> RateSourceBase:
        from hummingbot.data_feed.coin_gecko_data_feed.coin_gecko_constants import CoinGeckoAPITier
        try:
            api_tier_enum = CoinGeckoAPITier[api_tier.upper()]
        except KeyError:
            api_tier_enum = CoinGeckoAPITier.PUBLIC

        rate_source = RATE_ORACLE_SOURCES[cls.model_config["title"]](
            extra_token_ids=extra_tokens,
            api_key=api_key,
            api_tier=api_tier_enum,
        )
        rate_source.extra_token_ids = extra_tokens
        return rate_source


class CoinCapRateSourceMode(RateSourceModeBase):
    name: str = Field(default="coin_cap")
    assets_map: Dict[str, str] = Field(
        default=",".join(
            [
                ":".join(pair) for pair in {
                    "BTC": "bitcoin",
                    "ETH": "ethereum",
                    "USDT": "tether",
                    "CONV": "convergence",
                    "FIRO": "zcoin",
                    "BUSD": "binance-usd",
                    "ONE": "harmony",
                }.items()
            ]
        ),
        description=(
            "The symbol-to-asset ID map for CoinCap. Assets IDs can be found by selecting a symbol"
            " on https://coincap.io/ and extracting the last segment of the URL path."
        ),
        json_schema_extra={
            "prompt": lambda cm: (
                "CoinCap symbol-to-asset ID map (e.g. 'BTC:bitcoin,ETH:ethereum', find IDs on https://coincap.io/"
                " by selecting a symbol and extracting the last segment of the URL path)"
            ),
            "is_connect_key": True,
            "prompt_on_new": True,
        }
    )
    api_key: SecretStr = Field(
        default=SecretStr(""),
        description="API key to use to request information from CoinCap (if empty public requests will be used)",
        json_schema_extra={
            "prompt": lambda cm: "CoinCap API key (optional, but improves rate limits)",
            "is_secure": True,
            "is_connect_key": True,
            "prompt_on_new": True,
        }
    )
    model_config = ConfigDict(title="coin_cap")

    def build_rate_source(self) -> RateSourceBase:
        rate_source = RATE_ORACLE_SOURCES["coin_cap"](
            assets_map=self.assets_map, api_key=self.api_key.get_secret_value()
        )
        return rate_source

    @field_validator("assets_map", mode="before")
    @classmethod
    def validate_extra_tokens(cls, value: Union[str, Dict[str, str]]):
        if isinstance(value, str):
            value = {key: val for key, val in [v.split(":") for v in value.split(",")]}
        return value

    # === post-validations ===

    @model_validator(mode="after")
    def post_validations(self):
        RateOracle.get_instance().source = RATE_ORACLE_SOURCES["coin_cap"](
            assets_map=self.assets_map, api_key=self.api_key.get_secret_value())
        return self


class KuCoinRateSourceMode(ExchangeRateSourceModeBase):
    name: str = Field(default="kucoin")
    model_config = ConfigDict(title="kucoin")


class GateIoRateSourceMode(ExchangeRateSourceModeBase):
    name: str = Field(default="gate_io")
    model_config = ConfigDict(title="gate_io")


class DexalotRateSourceMode(ExchangeRateSourceModeBase):
    name: str = Field(default="dexalot")
    model_config = ConfigDict(title="dexalot")


class CoinbaseAdvancedTradeRateSourceMode(ExchangeRateSourceModeBase):
    name: str = Field(default="coinbase_advanced_trade")
    model_config = ConfigDict(title="coinbase_advanced_trade")
    use_auth_for_public_endpoints: bool = Field(
        default=False,
        description="Use authentication for public endpoints",
        json_schema_extra = {
            "prompt": lambda cm: "Would you like to use authentication for public endpoints? (Yes/No) (only affects rate limiting)",
            "prompt_on_new": True,
            "is_connect_key": True,
        },
    )

    def build_rate_source(self) -> RateSourceBase:
        return RATE_ORACLE_SOURCES[self.model_config["title"]](use_auth_for_public_endpoints=self.use_auth_for_public_endpoints)


class HyperliquidRateSourceMode(ExchangeRateSourceModeBase):
    name: str = Field(default="hyperliquid")
    model_config = ConfigDict(title="hyperliquid")


class HyperliquidPerpetualRateSourceMode(ExchangeRateSourceModeBase):
    name: str = Field(default="hyperliquid_perpetual")
    model_config = ConfigDict(title="hyperliquid_perpetual")


class DeriveRateSourceMode(ExchangeRateSourceModeBase):
    name: str = Field(default="derive")
    model_config = ConfigDict(title="derive")


RATE_SOURCE_MODES = {
    AscendExRateSourceMode.model_config["title"]: AscendExRateSourceMode,
    BinanceRateSourceMode.model_config["title"]: BinanceRateSourceMode,
    CoinGeckoRateSourceMode.model_config["title"]: CoinGeckoRateSourceMode,
    CoinCapRateSourceMode.model_config["title"]: CoinCapRateSourceMode,
    DexalotRateSourceMode.model_config["title"]: DexalotRateSourceMode,
    KuCoinRateSourceMode.model_config["title"]: KuCoinRateSourceMode,
    GateIoRateSourceMode.model_config["title"]: GateIoRateSourceMode,
    CoinbaseAdvancedTradeRateSourceMode.model_config["title"]: CoinbaseAdvancedTradeRateSourceMode,
    CubeRateSourceMode.model_config["title"]: CubeRateSourceMode,
    HyperliquidRateSourceMode.model_config["title"]: HyperliquidRateSourceMode,
    HyperliquidPerpetualRateSourceMode.model_config["title"]: HyperliquidPerpetualRateSourceMode,
    DeriveRateSourceMode.model_config["title"]: DeriveRateSourceMode,
    MexcRateSourceMode.model_config["title"]: MexcRateSourceMode,
}


class ClientConfigMap(BaseClientModel):
    instance_id: str = Field(
        default=generate_client_id(),
        json_schema_extra={"prompt": lambda cm: "Enter a unique identifier for this instance of Hummingbot"},
    )
    fetch_pairs_from_all_exchanges: bool = Field(
        default=False,
        description="Fetch trading pairs from all exchanges if True, otherwise fetch only from connected exchanges.",
        json_schema_extra={
            "prompt": lambda cm: "Would you like to fetch trading pairs from all exchanges? (True/False)"
        }
    )
    log_level: str = Field(default="INFO")
    debug_console: bool = Field(default=False)
    strategy_report_interval: float = Field(default=900)
    logger_override_whitelist: List = Field(
        default=["hummingbot.strategy.arbitrage", "hummingbot.strategy.cross_exchange_market_making", "conf"]
    )
    log_file_path: Path = Field(
        default=DEFAULT_LOG_FILE_PATH,
        json_schema_extra={"prompt": lambda cm: "Where would you like to save your logs? (default 'logs/hummingbot_logs.log')"},
    )
    kill_switch_mode: Union[tuple(KILL_SWITCH_MODES.values())] = Field(
        default=KillSwitchDisabledMode(),
        json_schema_extra={"prompt": lambda cm: f"Select the desired kill-switch mode ({'/'.join(list(KILL_SWITCH_MODES.keys()))})"},
    )
    autofill_import: AutofillImportEnum = Field(
        default=AutofillImportEnum.disabled,
        description="What to auto-fill in the prompt after each import command (start/config)",
        json_schema_extra={
            "prompt": lambda cm: f"What to auto-fill in the prompt after each import command? ({'/'.join(list(AutofillImportEnum))})"
        }
    )
    mqtt_bridge: MQTTBridgeConfigMap = Field(
        default=MQTTBridgeConfigMap(),
        description=('MQTT Bridge configuration.'),
    )
    send_error_logs: bool = Field(
        default=True,
        description="Error log sharing",
        json_schema_extra={"prompt": lambda cm: "Would you like to send error logs to hummingbot? (True/False)"},
    )
    db_mode: Union[tuple(DB_MODES.values())] = Field(
        default=DBSqliteMode(),
        description=("Advanced database options, currently supports SQLAlchemy's included dialects"
                     "\nReference: https://docs.sqlalchemy.org/en/13/dialects/"
                     "\nTo use an instance of SQLite DB the required configuration is \n  db_engine: sqlite"
                     "\nTo use a DBMS the required configuration is"
                     "\n  db_host: 127.0.0.1\n  db_port: 3306\n  db_username: username\n  db_password: password"
                     "\n  db_name: dbname"),
        json_schema_extra={"prompt": lambda cm: f"Select the desired db mode ({'/'.join(list(DB_MODES.keys()))})"},
    )
    balance_asset_limit: Dict[str, Dict[str, Decimal]] = Field(
        default={exchange: {} for exchange in AllConnectorSettings.get_exchange_names()},
        description=("Balance Limit Configurations"
                     "\ne.g. Setting USDT and BTC limits on Binance."
                     "\nbalance_asset_limit:"
                     "\n  binance:"
                     "\n    BTC: 0.1"
                     "\n    USDT: 1000"),
        json_schema_extra={"prompt": lambda cm: "Use the `balance limit` command e.g. balance limit [EXCHANGE] [ASSET] [AMOUNT]"},
    )
    manual_gas_price: Decimal = Field(
        default=Decimal("50"),
        description="Fixed gas price (in Gwei) for Ethereum transactions",
        gt=Decimal("0"),
        json_schema_extra={"prompt": lambda cm: "Enter fixed gas price (in Gwei) you want to use for Ethereum transactions"},
    )
    gateway: GatewayConfigMap = Field(
        default=GatewayConfigMap(),
        description=("Gateway API Configurations"
                     "\ndefault host to only use localhost"
                     "\nPort need to match the final installation port for Gateway"),
    )

    anonymized_metrics_mode: Union[tuple(METRICS_MODES.values())] = Field(
        default=AnonymizedMetricsEnabledMode(),
        description="Whether to enable aggregated order and trade data collection",
        json_schema_extra={"prompt": lambda cm: f"Select the desired metrics mode ({'/'.join(list(METRICS_MODES.keys()))})"},
    )
    rate_oracle_source: Union[tuple(RATE_SOURCE_MODES.values())] = Field(
        default=BinanceRateSourceMode(),
        description=f"A source for rate oracle, currently {', '.join(RATE_SOURCE_MODES.keys())}",
        json_schema_extra={"prompt": lambda cm: f"Select the desired rate oracle source ({'/'.join(RATE_SOURCE_MODES.keys())})"},
    )
    global_token: GlobalTokenConfigMap = Field(
        default=GlobalTokenConfigMap(),
        description="A universal token which to display tokens values in, e.g. USD,EUR,BTC"
    )
    rate_limits_share_pct: Decimal = Field(
        default=Decimal("100"),
        description=("Percentage of API rate limits (on any exchange and any end point) allocated to this bot instance."
                     "\nEnter 50 to indicate 50%. E.g. if the API rate limit is 100 calls per second, and you allocate "
                     "\n50% to this setting, the bot will have a maximum (limit) of 50 calls per second"),
        gt=Decimal("0"),
        le=Decimal("100"),
        json_schema_extra={"prompt": lambda cm: (
            "What percentage of API rate limits do you want to allocate to this bot instance?"
            " (Enter 50 to indicate 50%)"
        )},
    )
    commands_timeout: CommandsTimeoutConfigMap = Field(default=CommandsTimeoutConfigMap())
    tables_format: ClientConfigEnum(
        value="TabulateFormats",  # noqa: F821
        names={e: e for e in tabulate_formats},
        type=str,
    ) = Field(
        default="psql",
        description="Tabulate table format style (https://github.com/astanin/python-tabulate#table-format)",
        json_schema_extra={"prompt": lambda cm: (
            "What tabulate formatting to apply to the tables? [https://github.com/astanin/python-tabulate#table-format]"
        )}
    )
    paper_trade: PaperTradeConfigMap = Field(default=PaperTradeConfigMap())
    color: ColorConfigMap = Field(default=ColorConfigMap())
    tick_size: float = Field(
        default=1.0,
        ge=0.1,
        description="The tick size is the frequency with which the clock notifies the time iterators by calling the"
                    "\nc_tick() method, that means for example that if the tick size is 1, the logic of the strategy"
                    " \nwill run every second.",
        json_schema_extra={"prompt": lambda cm: (
            "What tick size (in seconds) do you want to use? (Enter 0.5 to indicate 0.5 seconds)"
        )},
    )
    market_data_collection: MarketDataCollectionConfigMap = Field(default=MarketDataCollectionConfigMap())
    model_config = ConfigDict(title="client_config_map")

    @field_validator("kill_switch_mode", mode="before")
    @classmethod
    def validate_kill_switch_mode(cls, v: Any):
        if isinstance(v, tuple(KILL_SWITCH_MODES.values())):
            return v  # Already a valid model

        if v == {}:
            return KillSwitchDisabledMode()

        if isinstance(v, dict):
            # Try validating against known mode models
            for mode_cls in KILL_SWITCH_MODES.values():
                try:
                    return mode_cls.model_validate(v)
                except Exception:
                    continue
            raise ValueError(f"Could not match dict to any known kill switch mode: {v}")

        if isinstance(v, str):
            if v not in KILL_SWITCH_MODES:
                raise ValueError(
                    f"Invalid kill switch mode string. Choose from: {list(KILL_SWITCH_MODES.keys())}."
                )
            return KILL_SWITCH_MODES[v].model_construct()

        raise ValueError(f"Unsupported type for kill switch mode: {type(v)}")

    @field_validator("autofill_import", mode="before")
    @classmethod
    def validate_autofill_import(cls, v: Union[str, AutofillImportEnum]):
        if isinstance(v, str) and v not in AutofillImportEnum.__members__:
            raise ValueError(f"The value must be one of {', '.join(list(AutofillImportEnum))}.")
        return v

    @field_validator("send_error_logs", "fetch_pairs_from_all_exchanges", mode="before")
    @classmethod
    def validate_bool(cls, v: str):
        """Used for client-friendly error output."""
        if isinstance(v, str):
            ret = validate_bool(v)
            if ret is not None:
                raise ValueError(ret)
        return v

    @field_validator("db_mode", mode="before")
    @classmethod
    def validate_db_mode(cls, v: Union[(str, Dict) + tuple(DB_MODES.values())]):
        if isinstance(v, tuple(DB_MODES.values()) + (Dict,)):
            sub_model = v
        elif v not in DB_MODES:
            raise ValueError(
                f"Invalid DB mode, please choose a value from {list(DB_MODES.keys())}."
            )
        else:
            sub_model = DB_MODES[v].model_construct()
        return sub_model

    @field_validator("anonymized_metrics_mode", mode="before")
    @classmethod
    def validate_anonymized_metrics_mode(cls, v: Union[(str, Dict) + tuple(METRICS_MODES.values())]):
        if isinstance(v, tuple(METRICS_MODES.values()) + (Dict,)):
            sub_model = v
        elif v not in METRICS_MODES:
            raise ValueError(
                f"Invalid metrics mode, please choose a value from {list(METRICS_MODES.keys())}."
            )
        else:
            sub_model = METRICS_MODES[v].model_construct()
        return sub_model

    @field_validator("rate_oracle_source", mode="before")
    @classmethod
    def validate_rate_oracle_source(cls, v: Any):
        if isinstance(v, tuple(RATE_SOURCE_MODES.values())):
            sub_model = v
        elif isinstance(v, dict):
            sub_model = RATE_SOURCE_MODES[v["name"]].model_construct()
        elif isinstance(v, str):
            sub_model = RATE_SOURCE_MODES[v].model_construct()
        elif v not in RATE_SOURCE_MODES:
            raise ValueError(
                f"Invalid rate source, please choose a value from {list(RATE_SOURCE_MODES.keys())}."
            )
        else:
            raise ValueError("Invalid rate source.")
        return sub_model

    @field_validator("tables_format", mode="before")
    @classmethod
    def validate_tables_format(cls, v: str):
        """Used for client-friendly error output."""
        if v not in tabulate_formats:
            raise ValueError("Invalid table format.")
        return v

    @field_validator("tick_size", mode="before")
    @classmethod
    def validate_tick_size(cls, v: float):
        """Used for client-friendly error output."""
        ret = validate_float(v, min_value=0.1)
        if ret is not None:
            raise ValueError(ret)
        return v

    # === post-validations ===

    @model_validator(mode="after")
    def post_validations(self):
        rate_source_mode: RateSourceModeBase = self.rate_oracle_source
        RateOracle.get_instance().source = rate_source_mode.build_rate_source()
        RateOracle.get_instance().quote_token = self.global_token.global_token_name
        return self



import asyncio
from decimal import Decimal
from typing import (
    Any,
    Dict,
    List,
    Optional,
)

from async_timeout import timeout

from hummingbot.core.data_type.common import OrderType, TradeType
from hummingbot.core.data_type.limit_order import LimitOrder

s_decimal_0 = Decimal(0)

GET_EX_ORDER_ID_TIMEOUT = 10  # seconds

cdef class InFlightOrderBase:
    def __init__(self,
                 client_order_id: str,
                 exchange_order_id: Optional[str],
                 trading_pair: str,
                 order_type: OrderType,
                 trade_type: TradeType,
                 price: Decimal,
                 amount: Decimal,
                 creation_timestamp: float,
                 initial_state: str):

        self.client_order_id = client_order_id
        self.exchange_order_id = exchange_order_id
        self.trading_pair = trading_pair
        self.order_type = order_type
        self.trade_type = trade_type
        self.price = price
        self.amount = amount
        self.executed_amount_base = s_decimal_0
        self.executed_amount_quote = s_decimal_0
        self.fee_asset = None
        self.fee_paid = s_decimal_0
        self.last_state = initial_state
        self.exchange_order_id_update_event = asyncio.Event()
        if self.exchange_order_id is not None:
            self.exchange_order_id_update_event.set()
        self.completely_filled_event = asyncio.Event()
        self._creation_timestamp = creation_timestamp

    def __repr__(self) -> str:
        return (f"InFlightOrder("
                f"client_order_id='{self.client_order_id}', "
                f"exchange_order_id='{self.exchange_order_id}', "
                f"creation_timestamp={self._creation_timestamp}, "
                f"trading_pair='{self.trading_pair}', "
                f"order_type={self.order_type}, "
                f"trade_type={self.trade_type}, "
                f"price={self.price}, "
                f"amount={self.amount}, "
                f"executed_amount_base={self.executed_amount_base}, "
                f"executed_amount_quote={self.executed_amount_quote}, "
                f"fee_asset='{self.fee_asset}', "
                f"fee_paid={self.fee_paid}, "
                f"last_state='{self.last_state}')")

    @property
    def is_done(self) -> bool:
        raise NotImplementedError

    @property
    def is_cancelled(self) -> bool:
        raise NotImplementedError

    @property
    def is_failure(self) -> bool:
        raise NotImplementedError

    @property
    def base_asset(self) -> str:
        return self.trading_pair.split("-")[0]

    @property
    def quote_asset(self) -> str:
        return self.trading_pair.split("-")[1]

    @property
    def creation_timestamp(self) -> float:
        """
        Returns the creation timestamp in seconds
        :return: The creation timestamp
        """
        if self._creation_timestamp > 0:
            timestamp = self._creation_timestamp
        else:
            timestamp = self._creation_timestamp_from_order_id()
        return timestamp

    def update_exchange_order_id(self, exchange_id: str):
        self.exchange_order_id = exchange_id
        self.exchange_order_id_update_event.set()

    async def get_exchange_order_id(self):
        if self.exchange_order_id is None:
            async with timeout(GET_EX_ORDER_ID_TIMEOUT):
                await self.exchange_order_id_update_event.wait()
        return self.exchange_order_id

    def to_limit_order(self) -> LimitOrder:
        return LimitOrder(
            self.client_order_id,
            self.trading_pair,
            self.trade_type is TradeType.BUY,
            self.base_asset,
            self.quote_asset,
            self.price,
            self.amount,
            creation_timestamp=int(self.creation_timestamp * 1e6)
        )

    def to_json(self) -> Dict[str, Any]:
        return {
            "client_order_id": self.client_order_id,
            "exchange_order_id": self.exchange_order_id,
            "trading_pair": self.trading_pair,
            "order_type": self.order_type.name,
            "trade_type": self.trade_type.name,
            "price": str(self.price),
            "amount": str(self.amount),
            "executed_amount_base": str(self.executed_amount_base),
            "executed_amount_quote": str(self.executed_amount_quote),
            "fee_asset": self.fee_asset,
            "fee_paid": str(self.fee_paid),
            "creation_timestamp": self.creation_timestamp,
            "last_state": self.last_state,
        }

    @classmethod
    def _instance_creation_parameters_from_json(cls, data: Dict[str, Any]) -> List[Any]:
        return [
            data["client_order_id"],
            data["exchange_order_id"],
            data["trading_pair"],
            getattr(OrderType, data["order_type"]),
            getattr(TradeType, data["trade_type"]),
            Decimal(data["price"]),
            Decimal(data["amount"]),
            data.get("creation_timestamp", -1),
            data["last_state"]]

    @classmethod
    def _basic_from_json(cls, data: Dict[str, Any]) -> InFlightOrderBase:
        """
        :param data: json data from API
        :return: formatted InFlightOrder
        """
        arguments = cls._instance_creation_parameters_from_json(data)
        order = cls(*arguments)
        order.executed_amount_base = Decimal(data["executed_amount_base"])
        order.executed_amount_quote = Decimal(data["executed_amount_quote"])
        order.fee_asset = data["fee_asset"]
        order.fee_paid = Decimal(data["fee_paid"])
        return order

    @classmethod
    def from_json(cls, data: Dict[str, Any]) -> InFlightOrderBase:
        """
        :param data: json data from API
        :return: formatted InFlightOrder
        """
        return cls._basic_from_json(data)

    def check_filled_condition(self):
        if (abs(self.amount) - self.executed_amount_base).quantize(Decimal('1e-8')) <= 0:
            self.completely_filled_event.set()

    async def wait_until_completely_filled(self):
        await self.completely_filled_event.wait()

    def _creation_timestamp_from_order_id(self) -> int:
        timestamp = -1
        if len(self.client_order_id) > 16:
            nonce_component = self.client_order_id[-16:]
            timestamp_string = f"{nonce_component[:10]}.{nonce_component[-6:]}"
            timestamp = float(timestamp_string) if nonce_component.isnumeric() else -1
        return timestamp



from bidict import bidict

from hummingbot.core.api_throttler.data_types import LinkedLimitWeightPair, RateLimit

REST_URL = "https://fapi.binance.com"
HEALTH_CHECK_ENDPOINT = "/fapi/v1/ping"
CANDLES_ENDPOINT = "/fapi/v1/klines"

WSS_URL = "wss://fstream.binance.com/ws"

INTERVALS = bidict({
    "1s": 1,
    "1m": 60,
    "3m": 180,
    "5m": 300,
    "15m": 900,
    "30m": 1800,
    "1h": 3600,
    "2h": 7200,
    "4h": 14400,
    "6h": 21600,
    "8h": 28800,
    "12h": 43200,
    "1d": 86400,
    "3d": 259200,
    "1w": 604800,
    "1M": 2592000
})
MAX_RESULTS_PER_CANDLESTICK_REST_REQUEST = 1500
REQUEST_WEIGHT = "REQUEST_WEIGHT"

RATE_LIMITS = [
    RateLimit(REQUEST_WEIGHT, limit=1200, time_interval=60),
    RateLimit(CANDLES_ENDPOINT, weight=2, limit=1200, time_interval=60, linked_limits=[LinkedLimitWeightPair("raw", 1)]),
    RateLimit(HEALTH_CHECK_ENDPOINT, limit=1200, time_interval=60, linked_limits=[LinkedLimitWeightPair("raw", 1)])]



import asyncio
import random
import time
from decimal import Decimal
from typing import Any, AsyncIterable, Dict, List, Optional, Tuple

from bidict import bidict

from hummingbot.connector.constants import s_decimal_NaN
from hummingbot.connector.derivative.aevo_perpetual import (
    aevo_perpetual_constants as CONSTANTS,
    aevo_perpetual_web_utils as web_utils,
)
from hummingbot.connector.derivative.aevo_perpetual.aevo_perpetual_api_order_book_data_source import (
    AevoPerpetualAPIOrderBookDataSource,
)
from hummingbot.connector.derivative.aevo_perpetual.aevo_perpetual_api_user_stream_data_source import (
    AevoPerpetualAPIUserStreamDataSource,
)
from hummingbot.connector.derivative.aevo_perpetual.aevo_perpetual_auth import AevoPerpetualAuth
from hummingbot.connector.derivative.position import Position
from hummingbot.connector.perpetual_derivative_py_base import PerpetualDerivativePyBase
from hummingbot.connector.trading_rule import TradingRule
from hummingbot.connector.utils import combine_to_hb_trading_pair, get_new_client_order_id
from hummingbot.core.api_throttler.data_types import RateLimit
from hummingbot.core.data_type.common import OrderType, PositionAction, PositionMode, PositionSide, PriceType, TradeType
from hummingbot.core.data_type.in_flight_order import InFlightOrder, OrderState, OrderUpdate, TradeUpdate
from hummingbot.core.data_type.order_book_tracker_data_source import OrderBookTrackerDataSource
from hummingbot.core.data_type.trade_fee import TokenAmount, TradeFeeBase
from hummingbot.core.data_type.user_stream_tracker_data_source import UserStreamTrackerDataSource
from hummingbot.core.utils.async_utils import safe_ensure_future
from hummingbot.core.utils.estimate_fee import build_trade_fee
from hummingbot.core.web_assistant.web_assistants_factory import WebAssistantsFactory


class AevoPerpetualDerivative(PerpetualDerivativePyBase):
    web_utils = web_utils

    SHORT_POLL_INTERVAL = 5.0
    LONG_POLL_INTERVAL = 120.0

    def __init__(
            self,
            balance_asset_limit: Optional[Dict[str, Dict[str, Decimal]]] = None,
            rate_limits_share_pct: Decimal = Decimal("100"),
            aevo_perpetual_api_key: str = None,
            aevo_perpetual_api_secret: str = None,
            aevo_perpetual_signing_key: str = None,
            aevo_perpetual_account_address: str = None,
            trading_pairs: Optional[List[str]] = None,
            trading_required: bool = True,
            domain: str = CONSTANTS.DEFAULT_DOMAIN,
    ):
        self._api_key = aevo_perpetual_api_key
        self._api_secret = aevo_perpetual_api_secret
        self._signing_key = aevo_perpetual_signing_key
        self._account_address = aevo_perpetual_account_address
        self._trading_required = trading_required
        self._trading_pairs = trading_pairs
        self._domain = domain
        self._position_mode = None
        self._last_trade_history_timestamp = None
        self._instrument_ids: Dict[str, int] = {}
        self._instrument_names: Dict[str, str] = {}
        super().__init__(balance_asset_limit, rate_limits_share_pct)

    @property
    def name(self) -> str:
        return self._domain

    @property
    def authenticator(self) -> Optional[AevoPerpetualAuth]:
        if self._api_key and self._api_secret and self._signing_key and self._account_address:
            return AevoPerpetualAuth(
                api_key=self._api_key,
                api_secret=self._api_secret,
                signing_key=self._signing_key,
                account_address=self._account_address,
                domain=self._domain,
            )
        return None

    @property
    def rate_limits_rules(self) -> List[RateLimit]:
        return CONSTANTS.RATE_LIMITS

    @property
    def domain(self) -> str:
        return self._domain

    @property
    def client_order_id_max_length(self) -> int:
        return CONSTANTS.MAX_ORDER_ID_LEN

    @property
    def client_order_id_prefix(self) -> str:
        return CONSTANTS.BROKER_ID

    @property
    def trading_rules_request_path(self) -> str:
        return CONSTANTS.MARKETS_PATH_URL

    @property
    def trading_pairs_request_path(self) -> str:
        return CONSTANTS.MARKETS_PATH_URL

    @property
    def check_network_request_path(self) -> str:
        return CONSTANTS.PING_PATH_URL

    @property
    def trading_pairs(self):
        return self._trading_pairs

    @property
    def is_cancel_request_in_exchange_synchronous(self) -> bool:
        return True

    @property
    def is_trading_required(self) -> bool:
        return self._trading_required

    @property
    def funding_fee_poll_interval(self) -> int:
        return 120

    @staticmethod
    def _signed_position_amount(amount: Decimal, position_side: PositionSide) -> Decimal:
        return -amount if position_side == PositionSide.SHORT else amount

    async def _make_network_check_request(self):
        await self._api_get(path_url=self.check_network_request_path)

    def get_price_by_type(self, trading_pair: str, price_type: PriceType) -> Decimal:
        price = super().get_price_by_type(trading_pair, price_type)
        if not price.is_nan():
            return price
        if price_type in {PriceType.MidPrice, PriceType.LastTrade}:
            fallback_price = self._get_funding_price_fallback(trading_pair)
            if fallback_price is not None:
                return fallback_price
        return price

    def supported_order_types(self) -> List[OrderType]:
        return [OrderType.LIMIT, OrderType.LIMIT_MAKER, OrderType.MARKET]

    async def get_all_pairs_prices(self) -> List[Dict[str, str]]:
        pairs_data = await self._api_get(
            path_url=CONSTANTS.MARKETS_PATH_URL,
            params={"instrument_type": CONSTANTS.PERPETUAL_INSTRUMENT_TYPE},
            limit_id=CONSTANTS.MARKETS_PATH_URL,
        )
        pairs_prices: List[Dict[str, str]] = []

        for pair_data in pairs_data:
            symbol = pair_data.get("instrument_name")
            price = pair_data.get("index_price")

            if symbol is None or price is None:
                continue

            pairs_prices.append({
                "symbol": symbol,
                "price": price,
            })

        return pairs_prices

    def supported_position_modes(self):
        return [PositionMode.ONEWAY]

    def set_position_mode(self, mode: PositionMode):
        if mode == PositionMode.HEDGE:
            self.logger().warning(
                "Aevo perpetual does not support HEDGE position mode. Using ONEWAY instead."
            )
            mode = PositionMode.ONEWAY
        super().set_position_mode(mode)

    def get_buy_collateral_token(self, trading_pair: str) -> str:
        trading_rule: TradingRule = self._trading_rules[trading_pair]
        return trading_rule.buy_order_collateral_token

    def get_sell_collateral_token(self, trading_pair: str) -> str:
        trading_rule: TradingRule = self._trading_rules[trading_pair]
        return trading_rule.sell_order_collateral_token

    def _is_request_exception_related_to_time_synchronizer(self, request_exception: Exception):
        return False

    def _is_order_not_found_during_status_update_error(self, status_update_exception: Exception) -> bool:
        error_message = str(status_update_exception)
        is_order_not_exist = CONSTANTS.NOT_EXIST_ERROR in error_message

        return is_order_not_exist

    def _is_order_not_found_during_cancelation_error(self, cancelation_exception: Exception) -> bool:
        error_message = str(cancelation_exception)
        is_order_not_exist = CONSTANTS.NOT_EXIST_ERROR in error_message

        return is_order_not_exist

    def _is_reduce_only_rejection_error(self, exception: Exception) -> bool:
        error_message = str(exception)
        return any(error_code in error_message for error_code in CONSTANTS.REDUCE_ONLY_REJECTION_ERRORS)

    def _on_order_failure(
        self,
        order_id: str,
        trading_pair: str,
        amount: Decimal,
        trade_type: TradeType,
        order_type: OrderType,
        price: Optional[Decimal],
        exception: Exception,
        **kwargs,
    ):
        position_action = kwargs.get("position_action")

        if position_action == PositionAction.CLOSE and self._is_reduce_only_rejection_error(exception):
            self.logger().info(
                f"Ignoring rejected reduce-only close order {order_id} ({trade_type.name} {trading_pair}): {exception}"
            )
            self._order_tracker.process_order_update(OrderUpdate(
                trading_pair=trading_pair,
                update_timestamp=self.current_timestamp,
                new_state=OrderState.CANCELED,
                client_order_id=order_id,
                misc_updates={
                    "error_message": str(exception),
                    "error_type": exception.__class__.__name__,
                },
            ))
            safe_ensure_future(self._update_positions())

            return

        super()._on_order_failure(
            order_id=order_id,
            trading_pair=trading_pair,
            amount=amount,
            trade_type=trade_type,
            order_type=order_type,
            price=price,
            exception=exception,
            **kwargs,
        )

    def _create_web_assistants_factory(self) -> WebAssistantsFactory:
        return web_utils.build_api_factory(
            throttler=self._throttler,
            auth=self._auth,
        )

    async def _make_trading_rules_request(self) -> Any:
        exchange_info = await self._api_get(
            path_url=self.trading_rules_request_path,
            params={"instrument_type": CONSTANTS.PERPETUAL_INSTRUMENT_TYPE},
        )
        return exchange_info

    async def _make_trading_pairs_request(self) -> Any:
        exchange_info = await self._api_get(
            path_url=self.trading_pairs_request_path,
            params={"instrument_type": CONSTANTS.PERPETUAL_INSTRUMENT_TYPE},
        )
        return exchange_info

    def _get_funding_price_fallback(self, trading_pair: str) -> Optional[Decimal]:
        try:
            funding_info = self.get_funding_info(trading_pair)
        except KeyError:
            return None
        price = funding_info.mark_price or funding_info.index_price or s_decimal_NaN
        return price if price > 0 else None

    def _resolve_trading_pair_symbols_duplicate(self, mapping: bidict, new_exchange_symbol: str, base: str, quote: str):
        expected_exchange_symbol = f"{base}{quote}"
        trading_pair = combine_to_hb_trading_pair(base, quote)
        current_exchange_symbol = mapping.inverse[trading_pair]
        if current_exchange_symbol == expected_exchange_symbol:
            pass
        elif new_exchange_symbol == expected_exchange_symbol:
            mapping.pop(current_exchange_symbol)
            mapping[new_exchange_symbol] = trading_pair
        else:
            self.logger().error(
                f"Could not resolve the exchange symbols {new_exchange_symbol} and {current_exchange_symbol}")
            mapping.pop(current_exchange_symbol)

    def _initialize_trading_pair_symbols_from_exchange_info(self, exchange_info: List):
        mapping = bidict()
        for symbol_data in filter(web_utils.is_exchange_information_valid, exchange_info):
            if symbol_data.get("instrument_type") != CONSTANTS.PERPETUAL_INSTRUMENT_TYPE:
                continue
            exchange_symbol = symbol_data["instrument_name"]
            base = symbol_data["underlying_asset"]
            quote = symbol_data["quote_asset"]
            trading_pair = combine_to_hb_trading_pair(base, quote)
            if trading_pair in mapping.inverse:
                self._resolve_trading_pair_symbols_duplicate(mapping, exchange_symbol, base, quote)
            else:
                mapping[exchange_symbol] = trading_pair
            self._instrument_ids[trading_pair] = int(symbol_data["instrument_id"])
            self._instrument_names[trading_pair] = exchange_symbol
        self._set_trading_pair_symbol_map(mapping)

    async def _format_trading_rules(self, exchange_info_dict: List) -> List[TradingRule]:
        return_val: List[TradingRule] = []
        for market in exchange_info_dict:
            try:
                if market.get("instrument_type") != CONSTANTS.PERPETUAL_INSTRUMENT_TYPE:
                    continue
                if not web_utils.is_exchange_information_valid(market):
                    continue

                base = market["underlying_asset"]
                quote = market["quote_asset"]
                trading_pair = combine_to_hb_trading_pair(base, quote)

                price_step = Decimal(str(market["price_step"]))
                amount_step = Decimal(str(market["amount_step"]))
                min_order_value = Decimal(str(market.get("min_order_value", "0")))

                return_val.append(
                    TradingRule(
                        trading_pair=trading_pair,
                        min_base_amount_increment=amount_step,
                        min_price_increment=price_step,
                        min_order_size=amount_step,
                        min_order_value=min_order_value,
                        buy_order_collateral_token=quote,
                        sell_order_collateral_token=quote,
                    )
                )
            except Exception:
                self.logger().error(f"Error parsing trading rule for {market}.", exc_info=True)
        return return_val

    def _create_order_book_data_source(self) -> OrderBookTrackerDataSource:
        return AevoPerpetualAPIOrderBookDataSource(
            trading_pairs=self._trading_pairs,
            connector=self,
            api_factory=self._web_assistants_factory,
            domain=self._domain,
        )

    def _create_user_stream_data_source(self) -> UserStreamTrackerDataSource:
        return AevoPerpetualAPIUserStreamDataSource(
            auth=self._auth,
            trading_pairs=self._trading_pairs,
            connector=self,
            api_factory=self._web_assistants_factory,
            domain=self._domain,
        )

    def _get_fee(self,
                 base_currency: str,
                 quote_currency: str,
                 order_type: OrderType,
                 order_side: TradeType,
                 position_action: PositionAction,
                 amount: Decimal,
                 price: Decimal = s_decimal_NaN,
                 is_maker: Optional[bool] = None) -> TradeFeeBase:
        is_maker = is_maker or False
        fee = build_trade_fee(
            self.name,
            is_maker,
            base_currency=base_currency,
            quote_currency=quote_currency,
            order_type=order_type,
            order_side=order_side,
            amount=amount,
            price=price,
        )
        return fee

    async def _update_trading_fees(self):
        """
        Update fees information from the exchange
        """
        pass

    def buy(self,
            trading_pair: str,
            amount: Decimal,
            order_type=OrderType.LIMIT,
            price: Decimal = s_decimal_NaN,
            **kwargs) -> str:
        order_id = get_new_client_order_id(
            is_buy=True,
            trading_pair=trading_pair,
            hbot_order_id_prefix=self.client_order_id_prefix,
            max_id_len=self.client_order_id_max_length,
        )
        if order_type is OrderType.MARKET:
            reference_price = self.get_mid_price(trading_pair) if price.is_nan() else price
            market_price = reference_price * (Decimal("1") + CONSTANTS.MARKET_ORDER_SLIPPAGE)
            price = self.quantize_order_price(trading_pair, market_price)

        safe_ensure_future(self._create_order(
            trade_type=TradeType.BUY,
            order_id=order_id,
            trading_pair=trading_pair,
            amount=amount,
            order_type=order_type,
            price=price,
            **kwargs))
        return order_id

    def sell(self,
             trading_pair: str,
             amount: Decimal,
             order_type: OrderType = OrderType.LIMIT,
             price: Decimal = s_decimal_NaN,
             **kwargs) -> str:
        order_id = get_new_client_order_id(
            is_buy=False,
            trading_pair=trading_pair,
            hbot_order_id_prefix=self.client_order_id_prefix,
            max_id_len=self.client_order_id_max_length,
        )
        if order_type is OrderType.MARKET:
            reference_price = self.get_mid_price(trading_pair) if price.is_nan() else price
            market_price = reference_price * (Decimal("1") - CONSTANTS.MARKET_ORDER_SLIPPAGE)
            price = self.quantize_order_price(trading_pair, market_price)

        safe_ensure_future(self._create_order(
            trade_type=TradeType.SELL,
            order_id=order_id,
            trading_pair=trading_pair,
            amount=amount,
            order_type=order_type,
            price=price,
            **kwargs))
        return order_id

    async def _place_order(
            self,
            order_id: str,
            trading_pair: str,
            amount: Decimal,
            trade_type: TradeType,
            order_type: OrderType,
            price: Decimal,
            position_action: PositionAction = PositionAction.NIL,
            **kwargs,
    ) -> Tuple[str, float]:

        instrument_id = self._instrument_ids.get(trading_pair)
        if instrument_id is None:
            self.logger().error(f"Order {order_id} rejected: instrument not found for {trading_pair}.")
            raise KeyError(f"Instrument not found for {trading_pair}")
        is_buy = trade_type is TradeType.BUY
        timestamp = int(time.time())
        salt = random.randint(0, 10 ** 6)
        limit_price = web_utils.decimal_to_int(price)
        amount_int = web_utils.decimal_to_int(amount)

        signature = self._auth.sign_order(
            is_buy=is_buy,
            limit_price=limit_price,
            amount=amount_int,
            salt=salt,
            instrument=instrument_id,
            timestamp=timestamp,
        )

        api_params = {
            "instrument": instrument_id,
            "maker": self._account_address,
            "is_buy": is_buy,
            "amount": str(amount_int),
            "limit_price": str(limit_price),
            "salt": str(salt),
            "signature": signature,
            "timestamp": str(timestamp),
            "post_only": order_type is OrderType.LIMIT_MAKER,
            "reduce_only": position_action is PositionAction.CLOSE,
            "time_in_force": "IOC" if order_type is OrderType.MARKET else "GTC",
        }
        order_result = await self._api_post(
            path_url=CONSTANTS.ORDERS_PATH_URL,
            data=api_params,
            is_auth_required=True,
            limit_id=CONSTANTS.ORDERS_PATH_URL,
        )
        if order_result.get("error") is not None:
            self.logger().error(f"Order {order_id} failed: {order_result['error']}")
            raise IOError(f"Error submitting order {order_id}: {order_result['error']}")

        exchange_order_id = str(order_result.get("order_id"))
        return exchange_order_id, self.current_timestamp

    async def _place_cancel(self, order_id: str, tracked_order: InFlightOrder):
        exchange_order_id = await tracked_order.get_exchange_order_id()
        if exchange_order_id is None:
            return False

        cancel_result = await self._api_delete(
            path_url=CONSTANTS.ORDER_PATH_URL.format(order_id=exchange_order_id),
            is_auth_required=True,
            limit_id=CONSTANTS.ORDERS_PATH_URL,
        )

        if cancel_result.get("error") is not None:
            raise IOError(f"{cancel_result['error']}")

        return True

    async def _request_order_status(self, tracked_order: InFlightOrder) -> OrderUpdate:
        exchange_order_id = await tracked_order.get_exchange_order_id()
        order_update = await self._api_get(
            path_url=CONSTANTS.ORDER_PATH_URL.format(order_id=exchange_order_id),
            is_auth_required=True,
            limit_id=CONSTANTS.ORDERS_PATH_URL,
        )
        if order_update.get("error") is not None:
            raise IOError(order_update["error"])
        current_state = order_update.get("order_status")
        update_timestamp = int(order_update.get("timestamp", order_update.get("created_timestamp", "0"))) * 1e-9
        return OrderUpdate(
            trading_pair=tracked_order.trading_pair,
            update_timestamp=update_timestamp,
            new_state=CONSTANTS.ORDER_STATE.get(current_state, OrderState.FAILED),
            client_order_id=tracked_order.client_order_id,
            exchange_order_id=str(order_update.get("order_id")),
        )

    async def _all_trade_updates_for_order(self, order: InFlightOrder) -> List[TradeUpdate]:
        exchange_order_id = str(order.exchange_order_id)
        if exchange_order_id is None:
            return []

        start_time = int(order.creation_timestamp * 1e9)
        response = await self._api_get(
            path_url=CONSTANTS.TRADE_HISTORY_PATH_URL,
            params={
                "start_time": start_time,
                "instrument_type": CONSTANTS.PERPETUAL_INSTRUMENT_TYPE,
                "limit": 50,
            },
            is_auth_required=True,
            limit_id=CONSTANTS.TRADE_HISTORY_PATH_URL,
        )
        trade_updates: List[TradeUpdate] = []
        for trade in response.get("trade_history", []):
            if str(trade.get("order_id")) != exchange_order_id:
                continue
            fee_asset = order.quote_asset
            position_action = order.position
            fee = TradeFeeBase.new_perpetual_fee(
                fee_schema=self.trade_fee_schema(),
                position_action=position_action,
                percent_token=fee_asset,
                flat_fees=[TokenAmount(amount=Decimal(trade["fees"]), token=fee_asset)],
            )
            trade_updates.append(TradeUpdate(
                trade_id=str(trade.get("trade_id")),
                client_order_id=order.client_order_id,
                exchange_order_id=exchange_order_id,
                trading_pair=order.trading_pair,
                fill_timestamp=int(trade.get("created_timestamp", "0")) * 1e-9,
                fill_price=Decimal(trade.get("price", "0")),
                fill_base_amount=Decimal(trade.get("amount", "0")),
                fill_quote_amount=Decimal(trade.get("price", "0")) * Decimal(trade.get("amount", "0")),
                fee=fee,
            ))
        return trade_updates

    async def _update_balances(self):
        account_info = await self._api_get(
            path_url=CONSTANTS.ACCOUNT_PATH_URL,
            is_auth_required=True,
            limit_id=CONSTANTS.ACCOUNT_PATH_URL,
        )
        balances = account_info.get("collaterals", [])
        if not balances and "collaterals" not in account_info:
            self.logger().warning(
                "Aevo account response did not include collaterals; balance update skipped.")
            return
        local_asset_names = set(self._account_balances.keys())
        remote_asset_names = set()

        for balance_entry in balances:
            asset_name = balance_entry["collateral_asset"]
            free_balance = Decimal(balance_entry.get("available_balance", "0"))
            total_balance = Decimal(balance_entry.get("balance", "0"))
            self._account_available_balances[asset_name] = free_balance
            self._account_balances[asset_name] = total_balance
            remote_asset_names.add(asset_name)

        asset_names_to_remove = local_asset_names.difference(remote_asset_names)
        for asset_name in asset_names_to_remove:
            del self._account_available_balances[asset_name]
            del self._account_balances[asset_name]

    async def _update_positions(self):
        positions_info = await self._api_get(
            path_url=CONSTANTS.POSITIONS_PATH_URL,
            is_auth_required=True,
            limit_id=CONSTANTS.POSITIONS_PATH_URL,
        )

        positions = positions_info.get("positions", [])
        active_pairs = set()

        for position in positions:
            if position.get("instrument_type") != CONSTANTS.PERPETUAL_INSTRUMENT_TYPE:
                continue

            trading_pair = await self.trading_pair_associated_to_exchange_symbol(position["instrument_name"])
            active_pairs.add(trading_pair)
            position_side = PositionSide.LONG if position.get("side") == "buy" else PositionSide.SHORT
            amount = self._signed_position_amount(
                amount=Decimal(position.get("amount", "0")),
                position_side=position_side,
            )
            entry_price = Decimal(position.get("avg_entry_price", "0"))
            unrealized_pnl = Decimal(position.get("unrealized_pnl", "0"))
            leverage = Decimal(position.get("leverage", "1"))
            pos_key = self._perpetual_trading.position_key(trading_pair, position_side)

            if amount != 0:
                self._perpetual_trading.set_position(
                    pos_key,
                    Position(
                        trading_pair=trading_pair,
                        position_side=position_side,
                        unrealized_pnl=unrealized_pnl,
                        entry_price=entry_price,
                        amount=amount,
                        leverage=leverage,
                    )
                )
            else:
                self._perpetual_trading.remove_position(pos_key)

        if not positions:
            keys = list(self._perpetual_trading.account_positions.keys())
            for key in keys:
                self._perpetual_trading.remove_position(key)

    async def _get_position_mode(self) -> Optional[PositionMode]:
        return PositionMode.ONEWAY

    async def _trading_pair_position_mode_set(self, mode: PositionMode, trading_pair: str) -> Tuple[bool, str]:
        return True, ""

    async def _ensure_instrument_id(self, trading_pair: str) -> bool:
        if trading_pair in self._instrument_ids:
            return True
        if not self.is_trading_required:
            return False
        try:
            await self._update_trading_rules()
        except Exception as exc:
            self.logger().network(
                f"Error updating trading rules while resolving instrument id for {trading_pair}: {exc}"
            )
        return trading_pair in self._instrument_ids

    async def _set_trading_pair_leverage(self, trading_pair: str, leverage: int) -> Tuple[bool, str]:
        if not await self._ensure_instrument_id(trading_pair):
            return False, "Instrument not found"
        instrument_id = self._instrument_ids.get(trading_pair)
        if instrument_id is None:
            return False, "Instrument not found"
        try:
            await self._api_post(
                path_url=CONSTANTS.ACCOUNT_LEVERAGE_PATH_URL,
                data={
                    "instrument": instrument_id,
                    "leverage": leverage,
                },
                is_auth_required=True,
                limit_id=CONSTANTS.ACCOUNT_LEVERAGE_PATH_URL,
            )
            self._perpetual_trading.set_leverage(trading_pair, leverage)
            return True, ""
        except Exception as exception:
            return False, f"Error setting leverage: {exception}"

    async def _fetch_last_fee_payment(self, trading_pair: str) -> Tuple[int, Decimal, Decimal]:
        return 0, Decimal("-1"), Decimal("-1")

    async def _user_stream_event_listener(self):
        user_channels = [
            CONSTANTS.WS_ORDERS_CHANNEL,
            CONSTANTS.WS_FILLS_CHANNEL,
            CONSTANTS.WS_POSITIONS_CHANNEL,
        ]
        async for event_message in self._iter_user_event_queue():
            try:
                if isinstance(event_message, dict):
                    channel: str = event_message.get("channel", None)
                    results = event_message.get("data", None)
                elif event_message is asyncio.CancelledError:
                    raise asyncio.CancelledError
                else:
                    raise Exception(event_message)

                if channel not in user_channels:
                    self.logger().error(
                        f"Unexpected message in user stream: {event_message}.")
                    continue

                if channel == CONSTANTS.WS_ORDERS_CHANNEL:
                    for order_msg in results.get("orders", []):
                        self._process_order_message(order_msg)
                elif channel == CONSTANTS.WS_FILLS_CHANNEL:
                    fill_msg = results.get("fill")
                    if fill_msg is not None:
                        await self._process_trade_message(fill_msg)
                elif channel == CONSTANTS.WS_POSITIONS_CHANNEL:
                    for position in results.get("positions", []):
                        await self._process_position_message(position)
            except asyncio.CancelledError:
                raise
            except Exception:
                self.logger().error(
                    "Unexpected error in user stream listener loop.", exc_info=True)
                await self._sleep(5.0)

    async def _process_position_message(self, position: Dict[str, Any]):
        if position.get("instrument_type") != CONSTANTS.PERPETUAL_INSTRUMENT_TYPE:
            return

        trading_pair = await self.trading_pair_associated_to_exchange_symbol(position["instrument_name"])
        position_side = PositionSide.LONG if position.get("side") == "buy" else PositionSide.SHORT
        amount = self._signed_position_amount(
            amount=Decimal(position.get("amount", "0")),
            position_side=position_side,
        )
        entry_price = Decimal(position.get("avg_entry_price", "0"))
        unrealized_pnl = Decimal(position.get("unrealized_pnl", "0"))
        leverage = Decimal(position.get("leverage", "1"))
        pos_key = self._perpetual_trading.position_key(trading_pair, position_side)

        if amount != 0:
            self._perpetual_trading.set_position(
                pos_key,
                Position(
                    trading_pair=trading_pair,
                    position_side=position_side,
                    unrealized_pnl=unrealized_pnl,
                    entry_price=entry_price,
                    amount=amount,
                    leverage=leverage,
                )
            )
        else:
            self._perpetual_trading.remove_position(pos_key)

    async def _process_trade_message(self, trade: Dict[str, Any]):
        exchange_order_id = str(trade.get("order_id", ""))
        tracked_order = self._order_tracker.all_fillable_orders_by_exchange_order_id.get(exchange_order_id)

        if tracked_order is None:
            all_orders = self._order_tracker.all_fillable_orders
            for _, order in all_orders.items():
                await order.get_exchange_order_id()
            tracked_order = self._order_tracker.all_fillable_orders_by_exchange_order_id.get(exchange_order_id)
            if tracked_order is None:
                self.logger().debug(
                    f"Ignoring trade message with id {exchange_order_id}: not in in_flight_orders.")
                return

        fee_asset = tracked_order.quote_asset
        fee = TradeFeeBase.new_perpetual_fee(
            fee_schema=self.trade_fee_schema(),
            position_action=tracked_order.position,
            percent_token=fee_asset,
            flat_fees=[TokenAmount(amount=Decimal(trade.get("fees", "0")), token=fee_asset)],
        )
        trade_update: TradeUpdate = TradeUpdate(
            trade_id=str(trade.get("trade_id")),
            client_order_id=tracked_order.client_order_id,
            exchange_order_id=exchange_order_id,
            trading_pair=tracked_order.trading_pair,
            fill_timestamp=int(trade.get("created_timestamp", "0")) * 1e-9,
            fill_price=Decimal(trade.get("price", "0")),
            fill_base_amount=Decimal(trade.get("filled", "0")),
            fill_quote_amount=Decimal(trade.get("price", "0")) * Decimal(trade.get("filled", "0")),
            fee=fee,
        )
        self._order_tracker.process_trade_update(trade_update)

    def _process_order_message(self, order_msg: Dict[str, Any]):
        exchange_order_id = str(order_msg.get("order_id", ""))
        tracked_order = self._order_tracker.all_updatable_orders_by_exchange_order_id.get(exchange_order_id)
        if not tracked_order:
            self.logger().debug(
                f"Ignoring order message with id {exchange_order_id}: not in in_flight_orders.")
            return
        current_state = order_msg.get("order_status")
        update_timestamp = int(order_msg.get("created_timestamp", "0")) * 1e-9
        order_update: OrderUpdate = OrderUpdate(
            trading_pair=tracked_order.trading_pair,
            update_timestamp=update_timestamp,
            new_state=CONSTANTS.ORDER_STATE.get(current_state, OrderState.FAILED),
            client_order_id=tracked_order.client_order_id,
            exchange_order_id=exchange_order_id,
        )
        self._order_tracker.process_order_update(order_update=order_update)

    async def _iter_user_event_queue(self) -> AsyncIterable[Dict[str, any]]:
        while True:
            try:
                yield await self._user_stream_tracker.user_stream.get()
            except asyncio.CancelledError:
                raise
            except Exception:
                self.logger().network(
                    "Unknown error. Retrying after 1 seconds.",
                    exc_info=True,
                    app_warning_msg="Could not fetch user events from Aevo. Check API key and network connection.",
                )
                await self._sleep(1.0)

    async def _get_last_traded_price(self, trading_pair: str) -> float:
        exchange_symbol = await self.exchange_symbol_associated_to_pair(trading_pair=trading_pair)
        response = await self._api_get(
            path_url=f"{CONSTANTS.INSTRUMENT_PATH_URL}/{exchange_symbol}",
            limit_id=CONSTANTS.INSTRUMENT_PATH_URL,
        )
        price = response.get("mark_price") or response.get("index_price") or "0"
        return float(price)



import asyncio
import logging
import time
from abc import ABC, abstractmethod
from decimal import Decimal
from typing import List, Tuple

from hummingbot.core.api_throttler.data_types import RateLimit, TaskLog
from hummingbot.logger.logger import HummingbotLogger

arc_logger = None
MAX_CAPACITY_REACHED_WARNING_INTERVAL = 30.0


class AsyncRequestContextBase(ABC):
    """
    An async context class ('async with' syntax) that checks for rate limit and waits for the capacity to be freed.
    It uses an async lock to prevent multiple instances of this class from accessing the `acquire()` function.
    """

    _last_max_cap_warning_ts: float = 0.0

    @classmethod
    def logger(cls) -> HummingbotLogger:
        global arc_logger
        if arc_logger is None:
            arc_logger = logging.getLogger(__name__)
        return arc_logger

    def __init__(self,
                 task_logs: List[TaskLog],
                 rate_limit: RateLimit,
                 related_limits: List[Tuple[RateLimit, int]],
                 lock: asyncio.Lock,
                 safety_margin_pct: float,
                 retry_interval: float = 0.1,
                 ):
        """
        Asynchronous context associated with each API request.
        :param task_logs: Shared task logs associated with this API request
        :param rate_limit: The RateLimit associated with this API Request
        :param related_limits: List of linked rate limits with its corresponding weight associated with this API Request
        :param lock: A shared asyncio.Lock used between all instances of APIRequestContextBase
        :param retry_interval: Time between each limit check
        """
        self._task_logs: List[TaskLog] = task_logs
        self._rate_limit: RateLimit = rate_limit
        self._related_limits: List[Tuple[RateLimit, int]] = related_limits
        self._lock: asyncio.Lock = lock
        self._safety_margin_pct: float = safety_margin_pct
        self._retry_interval: float = retry_interval

    def flush(self):
        """
        Remove task logs that have passed rate limit periods
        :return:
        """
        now: Decimal = Decimal(str(time.time()))
        self._task_logs[:] = [
            task for task in self._task_logs
            if now - Decimal(str(task.timestamp)) <= Decimal(str(task.rate_limit.time_interval * (1 + self._safety_margin_pct)))
        ]

    @abstractmethod
    def within_capacity(self) -> bool:
        raise NotImplementedError

    async def acquire(self):
        while True:
            async with self._lock:
                self.flush()

                if self.within_capacity():
                    break
            await asyncio.sleep(self._retry_interval)
        async with self._lock:
            now = time.time()
            # Each related limit is represented as it own individual TaskLog

            # Log the acquired rate limit into the tasks log
            new_logs = [
                TaskLog(timestamp=now, rate_limit=self._rate_limit, weight=self._rate_limit.weight)
            ] + [
                # Log its related limits into the tasks log as individual tasks
                TaskLog(timestamp=now, rate_limit=limit, weight=weight)
                for limit, weight in self._related_limits
            ]
            self._task_logs.extend(new_logs)

    async def __aenter__(self):
        await self.acquire()

    async def __aexit__(self, exc_type, exc, tb):
        pass



#!/usr/bin/env python

from collections import namedtuple
from datetime import datetime
from typing import List

import pandas as pd

from hummingbot.core.data_type.common import OrderType, TradeType
from hummingbot.core.data_type.trade_fee import TradeFeeBase


class Trade(namedtuple("_Trade", "trading_pair, side, price, amount, order_type, market, timestamp, trade_fee")):
    trading_pair: str
    side: TradeType
    price: float
    amount: float
    order_type: OrderType
    market: str
    timestamp: float
    trade_fee: TradeFeeBase

    @classmethod
    def to_pandas(cls, trades: List):
        columns: List[str] = ["trading_pair",
                              "price",
                              "quantity",
                              "order_type",
                              "trade_side",
                              "market",
                              "timestamp",
                              "fee_percent",
                              "flat_fee / gas"]
        data = []
        for trade in trades:
            if len(trade.trade_fee.flat_fees) == 0:
                flat_fee_str = "None"
            else:
                fee_strs = [f"{fee_tuple[0]} {fee_tuple[1]}" for fee_tuple in trade.trade_fee.flat_fees]
                flat_fee_str = ",".join(fee_strs)

            data.append([
                trade.trading_pair,
                trade.price,
                trade.amount,
                trade.order_type.name.lower(),
                trade.side.name.lower(),
                trade.market,
                datetime.fromtimestamp(trade.timestamp).strftime("%Y-%m-%d %H:%M:%S"),
                trade.trade_fee.percent,
                flat_fee_str,
            ])

        return pd.DataFrame(data=data, columns=columns)

    @property
    def trade_type(self):
        return self.side.name



from .event_listener cimport EventListener


cdef class EventLogger(EventListener):
    cdef:
        str _event_source
        object _logged_events
        object _generic_logged_events
        object _order_filled_logged_events
        dict _waiting
        dict _wait_returns
    cdef c_call(self, object event_object)



from abc import ABC, abstractmethod
from datetime import datetime, time
from typing import Union

from hummingbot.strategy.strategy_base import StrategyBase


class ConditionalExecutionState(ABC):
    """
    This class hierarchy models different execution conditions that can be used to alter the normal
    response from strategies to the tick (or c_tick) message.
    The default subclass is RunAlwaysExecutionState
    """

    _closing_time: int = None
    _time_left: int = None

    def __eq__(self, other):
        return type(self) is type(other)

    @property
    def time_left(self):
        return self._time_left

    @time_left.setter
    def time_left(self, value):
        self._time_left = value

    @property
    def closing_time(self):
        return self._closing_time

    @closing_time.setter
    def closing_time(self, value):
        self._closing_time = value

    @abstractmethod
    def process_tick(self, timestamp: float, strategy: StrategyBase):
        pass


class RunAlwaysExecutionState(ConditionalExecutionState):
    """
    Execution configuration to always run the strategy for every tick
    """

    def __str__(self):
        return "run continuously"

    def process_tick(self, timestamp: float, strategy: StrategyBase):
        self._closing_time = None
        self._time_left = None
        strategy.process_tick(timestamp)


class RunInTimeConditionalExecutionState(ConditionalExecutionState):
    """
    Execution configuration to always run the strategy only for the ticks that happen between the specified start
    timestamp and stop timestamp
    :param start_timestamp: Specifies the moment to start running the strategy (datetime or datetime.time)
    :param end_timestamp: Specifies the moment to stop running the strategy (datetime or datetime.time)
    """

    def __init__(self, start_timestamp: Union[datetime, time], end_timestamp: Union[datetime, time] = None):
        super().__init__()

        self._start_timestamp: Union[datetime, time] = start_timestamp
        self._end_timestamp: Union[datetime, time] = end_timestamp

    def __str__(self):
        if type(self._start_timestamp) is datetime:
            if self._end_timestamp is not None:
                return f"run between {self._start_timestamp} and {self._end_timestamp}"
            else:
                return f"run from {self._start_timestamp}"
        if type(self._start_timestamp) is time:
            if self._end_timestamp is not None:
                return f"run daily between {self._start_timestamp} and {self._end_timestamp}"

    def __eq__(self, other):
        return type(self) is type(other) and \
            self._start_timestamp == other._start_timestamp and \
            self._end_timestamp == other._end_timestamp

    def process_tick(self, timestamp: float, strategy: StrategyBase):
        if isinstance(self._start_timestamp, datetime):
            # From datetime
            # From datetime to datetime
            if self._end_timestamp is not None:

                self._closing_time = (self._end_timestamp.timestamp() - self._start_timestamp.timestamp()) * 1000

                if self._start_timestamp.timestamp() <= timestamp < self._end_timestamp.timestamp():
                    self._time_left = max((self._end_timestamp.timestamp() - timestamp) * 1000, 0)
                    strategy.process_tick(timestamp)
                else:
                    self._time_left = 0
                    strategy.cancel_active_orders()
                    strategy.logger().debug("Time span execution: tick will not be processed "
                                            f"(executing between {self._start_timestamp.isoformat(sep=' ')} "
                                            f"and {self._end_timestamp.isoformat(sep=' ')})")
            else:
                self._closing_time = None
                self._time_left = None
                if self._start_timestamp.timestamp() <= timestamp:
                    strategy.process_tick(timestamp)
                else:
                    strategy.cancel_active_orders()
                    strategy.logger().debug("Delayed start execution: tick will not be processed "
                                            f"(executing from {self._start_timestamp.isoformat(sep=' ')})")
        if isinstance(self._start_timestamp, time):
            # Daily between times
            if self._end_timestamp is not None:

                self._closing_time = (datetime.combine(datetime.today(), self._end_timestamp) - datetime.combine(datetime.today(), self._start_timestamp)).total_seconds() * 1000
                current_time = datetime.fromtimestamp(timestamp).time()

                if self._start_timestamp <= current_time < self._end_timestamp:
                    self._time_left = max((datetime.combine(datetime.today(), self._end_timestamp) - datetime.combine(datetime.today(), current_time)).total_seconds() * 1000, 0)
                    strategy.process_tick(timestamp)
                else:
                    self._time_left = 0
                    strategy.cancel_active_orders()
                    strategy.logger().debug("Time span execution: tick will not be processed "
                                            f"(executing between {self._start_timestamp} "
                                            f"and {self._end_timestamp})")



# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# NO CHECKED-IN PROTOBUF GENCODE
# source: PrivateDealsV3Api.proto
# Protobuf Python Version: 5.29.3
"""Generated protocol buffer code."""
from google.protobuf import (
    descriptor as _descriptor,
    descriptor_pool as _descriptor_pool,
    runtime_version as _runtime_version,
    symbol_database as _symbol_database,
)
from google.protobuf.internal import builder as _builder

_runtime_version.ValidateProtobufRuntimeVersion(
    _runtime_version.Domain.PUBLIC,
    5,
    29,
    3,
    '',
    'PrivateDealsV3Api.proto'
)
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()


DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(b'\n\x17PrivateDealsV3Api.proto\"\xec\x01\n\x11PrivateDealsV3Api\x12\r\n\x05price\x18\x01 \x01(\t\x12\x10\n\x08quantity\x18\x02 \x01(\t\x12\x0e\n\x06\x61mount\x18\x03 \x01(\t\x12\x11\n\ttradeType\x18\x04 \x01(\x05\x12\x0f\n\x07isMaker\x18\x05 \x01(\x08\x12\x13\n\x0bisSelfTrade\x18\x06 \x01(\x08\x12\x0f\n\x07tradeId\x18\x07 \x01(\t\x12\x15\n\rclientOrderId\x18\x08 \x01(\t\x12\x0f\n\x07orderId\x18\t \x01(\t\x12\x11\n\tfeeAmount\x18\n \x01(\t\x12\x13\n\x0b\x66\x65\x65\x43urrency\x18\x0b \x01(\t\x12\x0c\n\x04time\x18\x0c \x01(\x03\x42:\n\x1c\x63om.mxc.push.common.protobufB\x16PrivateDealsV3ApiProtoH\x01P\x01\x62\x06proto3')

_globals = globals()
_builder.BuildMessageAndEnumDescriptors(DESCRIPTOR, _globals)
_builder.BuildTopDescriptorsAndMessages(DESCRIPTOR, 'PrivateDealsV3Api_pb2', _globals)
if not _descriptor._USE_C_DESCRIPTORS:
    _globals['DESCRIPTOR']._loaded_options = None
    _globals['DESCRIPTOR']._serialized_options = b'\n\034com.mxc.push.common.protobufB\026PrivateDealsV3ApiProtoH\001P\001'
    _globals['_PRIVATEDEALSV3API']._serialized_start = 28
    _globals['_PRIVATEDEALSV3API']._serialized_end = 264
# @@protoc_insertion_point(module_scope)



import os
import sys
from os.path import dirname, join, realpath
from typing import Type

from prompt_toolkit.shortcuts import input_dialog, message_dialog
from prompt_toolkit.styles import Style

from hummingbot import root_path
from hummingbot.client.config.conf_migration import migrate_configs, migrate_non_secure_configs_only
from hummingbot.client.config.config_crypt import BaseSecretsManager, store_password_verification
from hummingbot.client.config.security import Security
from hummingbot.client.settings import CONF_DIR_PATH

sys.path.insert(0, str(root_path()))


with open(realpath(join(dirname(__file__), '../../VERSION'))) as version_file:
    version = version_file.read().strip()


def login_prompt(secrets_manager_cls: Type[BaseSecretsManager], style: Style):
    err_msg = None
    secrets_manager = None
    if Security.new_password_required():
        if legacy_confs_exist():
            secrets_manager = migrate_configs_prompt(secrets_manager_cls, style)
        else:
            show_welcome(style)
            password = input_dialog(
                title="Set Password",
                text="""
        Create a password to protect your sensitive data.
        This password is not shared with us nor with anyone else, so please store it securely.

        Enter your new password:""",
                password=True,
                style=style).run()
            if password is None:
                return None
            if password == str():
                err_msg = "The password must not be empty."
            else:
                re_password = input_dialog(
                    title="Set Password",
                    text="Please re-enter your password:",
                    password=True,
                    style=style).run()
                if re_password is None:
                    return None
                if password != re_password:
                    err_msg = "Passwords entered do not match, please try again."
                else:
                    secrets_manager = secrets_manager_cls(password)
                    store_password_verification(secrets_manager)
    else:
        password = input_dialog(
            title="Welcome back to Hummingbot",
            text="Enter your password:",
            password=True,
            style=style).run()
        if password is None:
            return None
        secrets_manager = secrets_manager_cls(password)
    if err_msg is None and not Security.login(secrets_manager):
        err_msg = "Invalid password - please try again."
    if err_msg is not None:
        message_dialog(
            title='Error',
            text=err_msg,
            style=style).run()
        return login_prompt(secrets_manager_cls, style)
    return secrets_manager


def legacy_confs_exist() -> bool:
    encrypted_conf_prefix = "encrypted_"
    encrypted_conf_postfix = ".json"
    exist = False
    for f in sorted(os.listdir(CONF_DIR_PATH)):
        f_path = CONF_DIR_PATH / f
        if os.path.isfile(f_path) and f.startswith(encrypted_conf_prefix) and f.endswith(encrypted_conf_postfix):
            exist = True
            break
    return exist


def migrate_configs_prompt(secrets_manager_cls: Type[BaseSecretsManager], style: Style) -> BaseSecretsManager:
    message_dialog(
        title='Configs Migration',
        text="""


            CONFIGS MIGRATION:

            We have recently refactored the way hummingbot handles configurations.
            To migrate your legacy configuration files to the new format,
            please enter your password on the following screen.

                """,
        style=style).run()
    password = input_dialog(
        title="Input Password",
        text="\n\nEnter your previous password:",
        password=True,
        style=style).run()
    if password is None:
        raise ValueError("Wrong password.")
    secrets_manager = secrets_manager_cls(password)
    errors = migrate_configs(secrets_manager)
    if len(errors) != 0:
        _migration_errors_dialog(errors, style)
    else:
        message_dialog(
            title='Configs Migration Success',
            text="""


                            CONFIGS MIGRATION SUCCESS:

                            The migration process was completed successfully.

                                """,
            style=style).run()
    return secrets_manager


def migrate_non_secure_only_prompt(style: Style):
    message_dialog(
        title='Configs Migration',
        text="""


                CONFIGS MIGRATION:

                We have recently refactored the way hummingbot handles configurations.
                We will now attempt to migrate any legacy config files to the new format.

                    """,
        style=style).run()
    errors = migrate_non_secure_configs_only()
    if len(errors) != 0:
        _migration_errors_dialog(errors, style)
    else:
        message_dialog(
            title='Configs Migration Success',
            text="""


                            CONFIGS MIGRATION SUCCESS:

                            The migration process was completed successfully.

                                """,
            style=style).run()


def _migration_errors_dialog(errors, style: Style):
    padding = "\n                    "
    errors_str = padding + padding.join(errors)
    message_dialog(
        title='Configs Migration Errors',
        text=f"""


                CONFIGS MIGRATION ERRORS:

                {errors_str}

                    """,
        style=style).run()


def show_welcome(style: Style):
    message_dialog(
        title='Welcome to Hummingbot',
        text="""

    ██╗  ██╗██╗   ██╗███╗   ███╗███╗   ███╗██╗███╗   ██╗ ██████╗ ██████╗  ██████╗ ████████╗
    ██║  ██║██║   ██║████╗ ████║████╗ ████║██║████╗  ██║██╔════╝ ██╔══██╗██╔═══██╗╚══██╔══╝
    ███████║██║   ██║██╔████╔██║██╔████╔██║██║██╔██╗ ██║██║  ███╗██████╔╝██║   ██║   ██║
    ██╔══██║██║   ██║██║╚██╔╝██║██║╚██╔╝██║██║██║╚██╗██║██║   ██║██╔══██╗██║   ██║   ██║
    ██║  ██║╚██████╔╝██║ ╚═╝ ██║██║ ╚═╝ ██║██║██║ ╚████║╚██████╔╝██████╔╝╚██████╔╝   ██║
    ╚═╝  ╚═╝ ╚═════╝ ╚═╝     ╚═╝╚═╝     ╚═╝╚═╝╚═╝  ╚═══╝ ╚═════╝ ╚═════╝  ╚═════╝    ╚═╝

    =======================================================================================

    Version: {version}
    Codebase: https://github.com/hummingbot/hummingbot


        """.format(version=version),
        style=style).run()
    message_dialog(
        title='Important Warning',
        text="""


    PLEASE READ THIS CAREFULLY BEFORE USING HUMMINGBOT:

    Hummingbot is a free and open source software client that helps you build algorithmic
    crypto trading strategies.

    Algorithmic crypto trading is a risky activity. You will be building a "bot" that
    automatically places orders and trades based on parameters that you set. Please take
    the time to understand how each strategy works before you risk real capital with it.
    You are solely responsible for the trades that you perform using Hummingbot.

        """,
        style=style).run()
    message_dialog(
        title='Important Warning',
        text="""


    SET A SECURE PASSWORD:

    To use Hummingbot, you will need to give it access to your crypto assets by entering
    your exchange API keys and/or wallet private keys. These keys are not shared with
    anyone, including us.

    On the next screen, you will set a password to protect these keys and other sensitive
    data. Please store this password safely since there is no way to reset it.

        """,
        style=style).run()



import asyncio
from typing import TYPE_CHECKING, Any, Dict, List, Optional

from hummingbot.connector.derivative.aevo_perpetual import (
    aevo_perpetual_constants as CONSTANTS,
    aevo_perpetual_web_utils as web_utils,
)
from hummingbot.connector.derivative.aevo_perpetual.aevo_perpetual_auth import AevoPerpetualAuth
from hummingbot.core.data_type.user_stream_tracker_data_source import UserStreamTrackerDataSource
from hummingbot.core.utils.async_utils import safe_ensure_future
from hummingbot.core.web_assistant.connections.data_types import WSJSONRequest, WSResponse
from hummingbot.core.web_assistant.web_assistants_factory import WebAssistantsFactory
from hummingbot.core.web_assistant.ws_assistant import WSAssistant
from hummingbot.logger import HummingbotLogger

if TYPE_CHECKING:
    from hummingbot.connector.derivative.aevo_perpetual.aevo_perpetual_derivative import AevoPerpetualDerivative


class AevoPerpetualAPIUserStreamDataSource(UserStreamTrackerDataSource):
    LISTEN_KEY_KEEP_ALIVE_INTERVAL = 1800
    WS_HEARTBEAT_TIME_INTERVAL = 30.0
    _logger: Optional[HummingbotLogger] = None

    def __init__(
            self,
            auth: AevoPerpetualAuth,
            trading_pairs: List[str],
            connector: 'AevoPerpetualDerivative',
            api_factory: WebAssistantsFactory,
            domain: str = CONSTANTS.DEFAULT_DOMAIN,
    ):
        super().__init__()
        self._domain = domain
        self._api_factory = api_factory
        self._auth = auth
        self._ws_assistants: List[WSAssistant] = []
        self._connector = connector
        self._trading_pairs: List[str] = trading_pairs

    @property
    def last_recv_time(self) -> float:
        if self._ws_assistant:
            return self._ws_assistant.last_recv_time
        return 0

    async def _get_ws_assistant(self) -> WSAssistant:
        if self._ws_assistant is None:
            self._ws_assistant = await self._api_factory.get_ws_assistant()
        return self._ws_assistant

    async def _authenticate(self, ws: WSAssistant):
        auth_payload = self._auth.get_ws_auth_payload()
        auth_request: WSJSONRequest = WSJSONRequest(payload=auth_payload)
        await ws.send(auth_request)
        response: WSResponse = await ws.receive()
        message = response.data
        if isinstance(message, dict) and message.get("error") is not None:
            raise IOError(f"Websocket authentication failed: {message['error']}")

    async def _connected_websocket_assistant(self) -> WSAssistant:
        ws: WSAssistant = await self._get_ws_assistant()
        url = f"{web_utils.wss_url(self._domain)}"
        await ws.connect(ws_url=url, ping_timeout=self.WS_HEARTBEAT_TIME_INTERVAL)
        safe_ensure_future(self._ping_thread(ws))
        return ws

    async def _subscribe_channels(self, websocket_assistant: WSAssistant):
        try:
            await self._authenticate(websocket_assistant)

            subscribe_payload = {
                "op": "subscribe",
                "data": [
                    CONSTANTS.WS_ORDERS_CHANNEL,
                    CONSTANTS.WS_FILLS_CHANNEL,
                    CONSTANTS.WS_POSITIONS_CHANNEL,
                ],
            }
            subscribe_request: WSJSONRequest = WSJSONRequest(payload=subscribe_payload)
            await websocket_assistant.send(subscribe_request)

            self.logger().info("Subscribed to private orders, fills and positions channels...")
        except asyncio.CancelledError:
            raise
        except Exception:
            self.logger().exception("Unexpected error occurred subscribing to user streams...")
            raise

    async def _process_event_message(self, event_message: Dict[str, Any], queue: asyncio.Queue):
        if event_message.get("error") is not None:
            err_msg = event_message.get("error", {}).get("message", event_message.get("error"))
            raise IOError({
                "label": "WSS_ERROR",
                "message": f"Error received via websocket - {err_msg}.",
            })
        if event_message.get("channel") in [
            CONSTANTS.WS_ORDERS_CHANNEL,
            CONSTANTS.WS_FILLS_CHANNEL,
            CONSTANTS.WS_POSITIONS_CHANNEL,
        ]:
            queue.put_nowait(event_message)

    async def _ping_thread(self, websocket_assistant: WSAssistant):
        try:
            ping_id = 1
            while True:
                ping_request = WSJSONRequest(payload={"op": "ping", "id": ping_id})
                await asyncio.sleep(self.WS_HEARTBEAT_TIME_INTERVAL)
                await websocket_assistant.send(ping_request)
                ping_id += 1
        except Exception as exc:
            self.logger().debug(f"Ping error {exc}")

    async def _process_websocket_messages(self, websocket_assistant: WSAssistant, queue: asyncio.Queue):
        while True:
            try:
                await super()._process_websocket_messages(
                    websocket_assistant=websocket_assistant,
                    queue=queue)
            except asyncio.TimeoutError:
                ping_request = WSJSONRequest(payload={"op": "ping", "id": 1})
                await websocket_assistant.send(ping_request)



import asyncio
import logging
from decimal import Decimal
from statistics import mean
from typing import Dict, List, Set, Union

import numpy as np
import pandas as pd

from hummingbot.connector.exchange_base import ExchangeBase
from hummingbot.connector.parrot import get_campaign_summary
from hummingbot.core.clock import Clock
from hummingbot.core.data_type.common import OrderType, TradeType
from hummingbot.core.data_type.limit_order import LimitOrder
from hummingbot.core.rate_oracle.rate_oracle import RateOracle
from hummingbot.core.utils.estimate_fee import build_trade_fee
from hummingbot.logger import HummingbotLogger
from hummingbot.strategy.market_trading_pair_tuple import MarketTradingPairTuple
from hummingbot.strategy.pure_market_making.inventory_skew_calculator import (
    calculate_bid_ask_ratios_from_base_asset_ratio,
)
from hummingbot.strategy.strategy_py_base import StrategyPyBase
from hummingbot.strategy.utils import order_age

from ...client.config.client_config_map import ClientConfigMap
from ...client.config.config_helpers import ClientConfigAdapter
from .data_types import PriceSize, Proposal

NaN = float("nan")
s_decimal_zero = Decimal(0)
s_decimal_nan = Decimal("NaN")
lms_logger = None


class LiquidityMiningStrategy(StrategyPyBase):

    @classmethod
    def logger(cls) -> HummingbotLogger:
        global lms_logger
        if lms_logger is None:
            lms_logger = logging.getLogger(__name__)
        return lms_logger

    def init_params(self,
                    client_config_map: Union[ClientConfigAdapter, ClientConfigMap],
                    exchange: ExchangeBase,
                    market_infos: Dict[str, MarketTradingPairTuple],
                    token: str,
                    order_amount: Decimal,
                    spread: Decimal,
                    inventory_skew_enabled: bool,
                    target_base_pct: Decimal,
                    order_refresh_time: float,
                    order_refresh_tolerance_pct: Decimal,
                    inventory_range_multiplier: Decimal = Decimal("1"),
                    volatility_interval: int = 60 * 5,
                    avg_volatility_period: int = 10,
                    volatility_to_spread_multiplier: Decimal = Decimal("1"),
                    max_spread: Decimal = Decimal("-1"),
                    max_order_age: float = 60. * 60.,
                    status_report_interval: float = 900,
                    hb_app_notification: bool = False):
        self._client_config_map = client_config_map
        self._exchange = exchange
        self._market_infos = market_infos
        self._empty_ob_market_infos = {}
        self._token = token
        self._order_amount = order_amount
        self._spread = spread
        self._order_refresh_time = order_refresh_time
        self._order_refresh_tolerance_pct = order_refresh_tolerance_pct
        self._inventory_skew_enabled = inventory_skew_enabled
        self._target_base_pct = target_base_pct
        self._inventory_range_multiplier = inventory_range_multiplier
        self._volatility_interval = volatility_interval
        self._avg_volatility_period = avg_volatility_period
        self._volatility_to_spread_multiplier = volatility_to_spread_multiplier
        self._max_spread = max_spread
        self._max_order_age = max_order_age
        self._ev_loop = asyncio.get_event_loop()
        self._last_timestamp = 0
        self._status_report_interval = status_report_interval
        self._ready_to_trade = False
        self._refresh_times = {market: 0 for market in market_infos}
        self._token_balances = {}
        self._sell_budgets = {}
        self._buy_budgets = {}
        self._mid_prices = {market: [] for market in market_infos}
        self._volatility = {market: s_decimal_nan for market in self._market_infos}
        self._last_vol_reported = 0.
        self._hb_app_notification = hb_app_notification

        self.add_markets([exchange])

    @property
    def active_orders(self):
        """
        List active orders (they have been sent to the market and have not been canceled yet)
        """
        limit_orders = self.order_tracker.active_limit_orders
        return [o[1] for o in limit_orders]

    @property
    def sell_budgets(self):
        return self._sell_budgets

    @property
    def buy_budgets(self):
        return self._buy_budgets

    def tick(self, timestamp: float):
        """
        Clock tick entry point, is run every second (on normal tick setting).
        :param timestamp: current tick timestamp
        """
        if not self._ready_to_trade:
            # Check if there are restored orders, they should be canceled before strategy starts.
            self._ready_to_trade = self._exchange.ready and len(self._exchange.limit_orders) == 0
            if not self._exchange.ready:
                self.logger().warning(f"{self._exchange.name} is not ready. Please wait...")
                return
            else:
                self.logger().info(f"{self._exchange.name} is ready. Trading started.")
                if self._validate_order_book_for_markets() >= 1:
                    self.create_budget_allocation()
                else:
                    self.logger().warning(f"{self._exchange.name} has no pairs with order book. Consider redefining your strategy.")
                    return

        self.update_mid_prices()
        self.update_volatility()
        proposals = self.create_base_proposals()
        self._token_balances = self.adjusted_available_balances()
        if self._inventory_skew_enabled:
            self.apply_inventory_skew(proposals)
        self.apply_budget_constraint(proposals)
        self.cancel_active_orders(proposals)
        self.execute_orders_proposal(proposals)

        self._last_timestamp = timestamp

    async def active_orders_df(self) -> pd.DataFrame:
        """
        Return the active orders in a DataFrame.
        """
        size_q_col = f"Amt({self._token})" if self.is_token_a_quote_token() else "Amt(Quote)"
        columns = ["Market", "Side", "Price", "Spread", "Amount", size_q_col, "Age"]
        data = []
        for order in self.active_orders:
            mid_price = self._market_infos[order.trading_pair].get_mid_price()
            spread = 0 if mid_price == 0 else abs(order.price - mid_price) / mid_price
            size_q = order.quantity * mid_price
            age = order_age(order, self.current_timestamp)
            # // indicates order is a paper order so 'n/a'. For real orders, calculate age.
            age_txt = "n/a" if age <= 0. else pd.Timestamp(age, unit='s').strftime('%H:%M:%S')
            data.append([
                order.trading_pair,
                "buy" if order.is_buy else "sell",
                float(order.price),
                f"{spread:.2%}",
                float(order.quantity),
                float(size_q),
                age_txt
            ])
        df = pd.DataFrame(data=data, columns=columns)
        df.sort_values(by=["Market", "Side"], inplace=True)
        return df

    def budget_status_df(self) -> pd.DataFrame:
        """
        Return the trader's budget in a DataFrame
        """
        data = []
        columns = ["Market", f"Budget({self._token})", "Base bal", "Quote bal", "Base/Quote"]
        for market, market_info in self._market_infos.items():
            mid_price = market_info.get_mid_price()
            base_bal = self._sell_budgets[market]
            quote_bal = self._buy_budgets[market]
            total_bal_in_quote = (base_bal * mid_price) + quote_bal
            total_bal_in_token = total_bal_in_quote
            if not self.is_token_a_quote_token():
                total_bal_in_token = base_bal + (quote_bal / mid_price)
            base_pct = (base_bal * mid_price) / total_bal_in_quote if total_bal_in_quote > 0 else s_decimal_zero
            quote_pct = quote_bal / total_bal_in_quote if total_bal_in_quote > 0 else s_decimal_zero
            data.append([
                market,
                float(total_bal_in_token),
                float(base_bal),
                float(quote_bal),
                f"{base_pct:.0%} / {quote_pct:.0%}"
            ])
        df = pd.DataFrame(data=data, columns=columns).replace(np.nan, '', regex=True)
        df.sort_values(by=["Market"], inplace=True)
        return df

    def market_status_df(self) -> pd.DataFrame:
        """
        Return the market status (prices, volatility) in a DataFrame
        """
        data = []
        columns = ["Market", "Mid price", "Best bid", "Best ask", "Volatility"]
        for market, market_info in self._market_infos.items():
            mid_price = market_info.get_mid_price()
            best_bid = self._exchange.get_price(market, False)
            best_ask = self._exchange.get_price(market, True)
            best_bid_pct = abs(best_bid - mid_price) / mid_price
            best_ask_pct = (best_ask - mid_price) / mid_price
            data.append([
                market,
                float(mid_price),
                f"{best_bid_pct:.2%}",
                f"{best_ask_pct:.2%}",
                "" if self._volatility[market].is_nan() else f"{self._volatility[market]:.2%}",
            ])
        df = pd.DataFrame(data=data, columns=columns).replace(np.nan, '', regex=True)
        df.sort_values(by=["Market"], inplace=True)
        return df

    async def miner_status_df(self) -> pd.DataFrame:
        """
        Return the miner status (payouts, rewards, liquidity, etc.) in a DataFrame
        """
        data = []
        g_sym = self._client_config_map.global_token.global_token_symbol
        columns = ["Market", "Payout", "Reward/wk", "Liquidity", "Yield/yr", "Max spread"]
        campaigns = await get_campaign_summary(self._exchange.display_name, list(self._market_infos.keys()))
        for market, campaign in campaigns.items():
            reward = await RateOracle.get_instance().get_value(
                amount=campaign.reward_per_wk, base_token=campaign.payout_asset
            )
            data.append([
                market,
                campaign.payout_asset,
                f"{g_sym}{reward:.0f}",
                f"{g_sym}{campaign.liquidity_usd:.0f}",
                f"{campaign.apy:.2%}",
                f"{campaign.spread_max:.2%}%"
            ])
        df = pd.DataFrame(data=data, columns=columns).replace(np.nan, '', regex=True)
        df.sort_values(by=["Market"], inplace=True)
        return df

    async def format_status(self) -> str:
        """
        Return the budget, market, miner and order statuses.
        """
        if not self._ready_to_trade:
            return "Market connectors are not ready."
        lines = []
        warning_lines = []
        warning_lines.extend(self.network_warning(list(self._market_infos.values())))

        budget_df = self.budget_status_df()
        lines.extend(["", "  Budget:"] + ["    " + line for line in budget_df.to_string(index=False).split("\n")])

        market_df = self.market_status_df()
        lines.extend(["", "  Markets:"] + ["    " + line for line in market_df.to_string(index=False).split("\n")])

        miner_df = await self.miner_status_df()
        if not miner_df.empty:
            lines.extend(["", "  Miner:"] + ["    " + line for line in miner_df.to_string(index=False).split("\n")])

        # See if there are any open orders.
        if len(self.active_orders) > 0:
            df = await self.active_orders_df()
            lines.extend(["", "  Orders:"] + ["    " + line for line in df.to_string(index=False).split("\n")])
        else:
            lines.extend(["", "  No active maker orders."])

        warning_lines.extend(self.balance_warning(list(self._market_infos.values())))
        if len(warning_lines) > 0:
            lines.extend(["", "*** WARNINGS ***"] + warning_lines)
        return "\n".join(lines)

    def start(self, clock: Clock, timestamp: float):
        restored_orders = self._exchange.limit_orders
        for order in restored_orders:
            self._exchange.cancel(order.trading_pair, order.client_order_id)

    def stop(self, clock: Clock):
        pass

    def _validate_order_book_for_markets(self):
        """
        Verify that the active markets have non-empty order books. Put the trading on hold
        for pairs with empty order books, in case it is a glitch?
        """
        markets = self._market_infos.copy()
        markets.update(self._empty_ob_market_infos)
        for market, market_info in markets.items():
            if markets[market].get_price(False).is_nan():
                self._empty_ob_market_infos[market] = market_info
                if market in self._market_infos:
                    self.logger().warning(f"{market} has an empty order book. Trading is paused")
                    self._market_infos.pop(market)
            else:
                self._market_infos[market] = market_info
                if market in self._empty_ob_market_infos:
                    self.logger().warning(f"{market} is being reactivated")
                    self._empty_ob_market_infos.pop(market)
        return len(self._market_infos)

    def create_base_proposals(self):
        """
        Each tick this strategy creates a set of proposals based on the market_info and the parameters from the
        constructor.
        """
        proposals = []
        for market, market_info in self._market_infos.items():
            spread = self._spread
            if not self._volatility[market].is_nan():
                # volatility applies only when it is higher than the spread setting.
                spread = max(spread, self._volatility[market] * self._volatility_to_spread_multiplier)
            if self._max_spread > s_decimal_zero:
                spread = min(spread, self._max_spread)
            mid_price = market_info.get_mid_price()
            buy_price = mid_price * (Decimal("1") - spread)
            buy_price = self._exchange.quantize_order_price(market, buy_price)
            buy_size = self.base_order_size(market, buy_price)
            sell_price = mid_price * (Decimal("1") + spread)
            sell_price = self._exchange.quantize_order_price(market, sell_price)
            sell_size = self.base_order_size(market, sell_price)
            proposals.append(Proposal(market, PriceSize(buy_price, buy_size), PriceSize(sell_price, sell_size)))
        return proposals

    def total_port_value_in_token(self) -> Decimal:
        """
        Total portfolio value in self._token amount
        """
        all_bals = self.adjusted_available_balances()
        port_value = all_bals.get(self._token, s_decimal_zero)
        for market, market_info in self._market_infos.items():
            base, quote = market.split("-")
            if self.is_token_a_quote_token():
                port_value += all_bals[base] * market_info.get_mid_price()
            else:
                port_value += all_bals[quote] / market_info.get_mid_price()
        return port_value

    def create_budget_allocation(self):
        """
        Create buy and sell budgets for every market
        """
        self._sell_budgets = {m: s_decimal_zero for m in self._market_infos}
        self._buy_budgets = {m: s_decimal_zero for m in self._market_infos}
        portfolio_value = self.total_port_value_in_token()
        market_portion = portfolio_value / len(self._market_infos)
        balances = self.adjusted_available_balances()
        for market, market_info in self._market_infos.items():
            base, quote = market.split("-")
            if self.is_token_a_quote_token():
                self._sell_budgets[market] = balances[base]
                buy_budget = market_portion - (balances[base] * market_info.get_mid_price())
                if buy_budget > s_decimal_zero:
                    self._buy_budgets[market] = buy_budget
            else:
                self._buy_budgets[market] = balances[quote]
                sell_budget = market_portion - (balances[quote] / market_info.get_mid_price())
                if sell_budget > s_decimal_zero:
                    self._sell_budgets[market] = sell_budget

    def base_order_size(self, trading_pair: str, price: Decimal = s_decimal_zero):
        base, quote = trading_pair.split("-")
        if self._token == base:
            return self._order_amount
        if price == s_decimal_zero:
            price = self._market_infos[trading_pair].get_mid_price()
        return self._order_amount / price

    def apply_budget_constraint(self, proposals: List[Proposal]):
        balances = self._token_balances.copy()
        for proposal in proposals:
            if balances[proposal.base()] < proposal.sell.size:
                proposal.sell.size = balances[proposal.base()]
            proposal.sell.size = self._exchange.quantize_order_amount(proposal.market, proposal.sell.size)
            balances[proposal.base()] -= proposal.sell.size

            quote_size = proposal.buy.size * proposal.buy.price
            quote_size = balances[proposal.quote()] if balances[proposal.quote()] < quote_size else quote_size
            buy_fee = build_trade_fee(self._exchange.name, True, proposal.base(), proposal.quote(),
                                      OrderType.LIMIT, TradeType.BUY, proposal.buy.size, proposal.buy.price)
            buy_size = quote_size / (proposal.buy.price * (Decimal("1") + buy_fee.percent))
            proposal.buy.size = self._exchange.quantize_order_amount(proposal.market, buy_size)
            balances[proposal.quote()] -= quote_size

    def is_within_tolerance(self, cur_orders: List[LimitOrder], proposal: Proposal):
        """
        False if there are no buys or sells or if the difference between the proposed price and current price is less
        than the tolerance. The tolerance value is strict max, cannot be equal.
        """
        cur_buy = [o for o in cur_orders if o.is_buy]
        cur_sell = [o for o in cur_orders if not o.is_buy]
        if (cur_buy and proposal.buy.size <= 0) or (cur_sell and proposal.sell.size <= 0):
            return False
        if cur_buy and \
                abs(proposal.buy.price - cur_buy[0].price) / cur_buy[0].price > self._order_refresh_tolerance_pct:
            return False
        if cur_sell and \
                abs(proposal.sell.price - cur_sell[0].price) / cur_sell[0].price > self._order_refresh_tolerance_pct:
            return False
        return True

    def cancel_active_orders(self, proposals: List[Proposal]):
        """
        Cancel any orders that have an order age greater than self._max_order_age or if orders are not within tolerance
        """
        for proposal in proposals:
            to_cancel = False
            cur_orders = [o for o in self.active_orders if o.trading_pair == proposal.market]
            if cur_orders and any(order_age(o, self.current_timestamp) > self._max_order_age for o in cur_orders):
                to_cancel = True
            elif self._refresh_times[proposal.market] <= self.current_timestamp and \
                    cur_orders and not self.is_within_tolerance(cur_orders, proposal):
                to_cancel = True
            if to_cancel:
                for order in cur_orders:
                    self.cancel_order(self._market_infos[proposal.market], order.client_order_id)
                    # To place new order on the next tick
                    self._refresh_times[order.trading_pair] = self.current_timestamp + 0.1

    def execute_orders_proposal(self, proposals: List[Proposal]):
        """
        Execute a list of proposals if the current timestamp is less than its refresh timestamp.
        Update the refresh timestamp.
        """
        for proposal in proposals:
            maker_order_type: OrderType = self._exchange.get_maker_order_type()
            cur_orders = [o for o in self.active_orders if o.trading_pair == proposal.market]
            if cur_orders or self._refresh_times[proposal.market] > self.current_timestamp:
                continue
            mid_price = self._market_infos[proposal.market].get_mid_price()
            spread = s_decimal_zero
            if proposal.buy.size > 0:
                spread = abs(proposal.buy.price - mid_price) / mid_price
                self.logger().info(f"({proposal.market}) Creating a bid order {proposal.buy} value: "
                                   f"{proposal.buy.size * proposal.buy.price:.2f} {proposal.quote()} spread: "
                                   f"{spread:.2%}")
                self.buy_with_specific_market(
                    self._market_infos[proposal.market],
                    proposal.buy.size,
                    order_type=maker_order_type,
                    price=proposal.buy.price
                )
            if proposal.sell.size > 0:
                spread = abs(proposal.sell.price - mid_price) / mid_price
                self.logger().info(f"({proposal.market}) Creating an ask order at {proposal.sell} value: "
                                   f"{proposal.sell.size * proposal.sell.price:.2f} {proposal.quote()} spread: "
                                   f"{spread:.2%}")
                self.sell_with_specific_market(
                    self._market_infos[proposal.market],
                    proposal.sell.size,
                    order_type=maker_order_type,
                    price=proposal.sell.price
                )
            if proposal.buy.size > 0 or proposal.sell.size > 0:
                if not self._volatility[proposal.market].is_nan() and spread > self._spread:
                    adjusted_vol = self._volatility[proposal.market] * self._volatility_to_spread_multiplier
                    if adjusted_vol > self._spread:
                        self.logger().info(f"({proposal.market}) Spread is widened to {spread:.2%} due to high "
                                           f"market volatility")

                self._refresh_times[proposal.market] = self.current_timestamp + self._order_refresh_time

    def is_token_a_quote_token(self):
        """
        Check if self._token is a quote token
        """
        quotes = self.all_quote_tokens()
        if len(quotes) == 1 and self._token in quotes:
            return True
        return False

    def all_base_tokens(self) -> Set[str]:
        """
        Get the base token (left-hand side) from all markets in this strategy
        """
        tokens = set()
        for market in self._market_infos:
            tokens.add(market.split("-")[0])
        return tokens

    def all_quote_tokens(self) -> Set[str]:
        """
        Get the quote token (right-hand side) from all markets in this strategy
        """
        tokens = set()
        for market in self._market_infos:
            tokens.add(market.split("-")[1])
        return tokens

    def all_tokens(self) -> Set[str]:
        """
        Return a list of all tokens involved in this strategy (base and quote)
        """
        tokens = set()
        for market in self._market_infos:
            tokens.update(market.split("-"))
        return tokens

    def adjusted_available_balances(self) -> Dict[str, Decimal]:
        """
        Calculates all available balances, account for amount attributed to orders and reserved balance.
        :return: a dictionary of token and its available balance
        """
        tokens = self.all_tokens()
        adjusted_bals = {t: s_decimal_zero for t in tokens}
        total_bals = {t: s_decimal_zero for t in tokens}
        total_bals.update(self._exchange.get_all_balances())
        for token in tokens:
            adjusted_bals[token] = self._exchange.get_available_balance(token)
        for order in self.active_orders:
            base, quote = order.trading_pair.split("-")
            if order.is_buy:
                adjusted_bals[quote] += order.quantity * order.price
            else:
                adjusted_bals[base] += order.quantity
        return adjusted_bals

    def apply_inventory_skew(self, proposals: List[Proposal]):
        """
        Apply an inventory split between the quote and base asset
        """
        for proposal in proposals:
            buy_budget = self._buy_budgets[proposal.market]
            sell_budget = self._sell_budgets[proposal.market]
            mid_price = self._market_infos[proposal.market].get_mid_price()
            total_order_size = proposal.sell.size + proposal.buy.size
            bid_ask_ratios = calculate_bid_ask_ratios_from_base_asset_ratio(
                float(sell_budget),
                float(buy_budget),
                float(mid_price),
                float(self._target_base_pct),
                float(total_order_size * self._inventory_range_multiplier)
            )
            proposal.buy.size *= Decimal(bid_ask_ratios.bid_ratio)
            proposal.sell.size *= Decimal(bid_ask_ratios.ask_ratio)

    def did_fill_order(self, event):
        """
        Check if order has been completed, log it, notify the hummingbot application, and update budgets.
        """
        order_id = event.order_id
        market_info = self.order_tracker.get_shadow_market_pair_from_order_id(order_id)
        if market_info is not None:
            if event.trade_type is TradeType.BUY:
                msg = f"({market_info.trading_pair}) Maker BUY order (price: {event.price}) of {event.amount} " \
                      f"{market_info.base_asset} is filled."
                self.log_with_clock(logging.INFO, msg)
                self.notify_hb_app_with_timestamp(msg)
                self._buy_budgets[market_info.trading_pair] -= (event.amount * event.price)
                self._sell_budgets[market_info.trading_pair] += event.amount
            else:
                msg = f"({market_info.trading_pair}) Maker SELL order (price: {event.price}) of {event.amount} " \
                      f"{market_info.base_asset} is filled."
                self.log_with_clock(logging.INFO, msg)
                self.notify_hb_app_with_timestamp(msg)
                self._sell_budgets[market_info.trading_pair] -= event.amount
                self._buy_budgets[market_info.trading_pair] += (event.amount * event.price)

    def update_mid_prices(self):
        """
        Query asset markets for mid price
        """
        for market in self._market_infos:
            mid_price = self._market_infos[market].get_mid_price()
            self._mid_prices[market].append(mid_price)
            # To avoid memory leak, we store only the last part of the list needed for volatility calculation
            max_len = self._volatility_interval * self._avg_volatility_period
            self._mid_prices[market] = self._mid_prices[market][-1 * max_len:]

    def update_volatility(self):
        """
        Update volatility data from the market
        """
        self._volatility = {market: s_decimal_nan for market in self._market_infos}
        for market, mid_prices in self._mid_prices.items():
            last_index = len(mid_prices) - 1
            atr = []
            first_index = last_index - (self._volatility_interval * self._avg_volatility_period)
            first_index = max(first_index, 0)
            for i in range(last_index, first_index, self._volatility_interval * -1):
                prices = mid_prices[i - self._volatility_interval + 1: i + 1]
                if not prices:
                    break
                atr.append((max(prices) - min(prices)) / min(prices))
            if atr:
                self._volatility[market] = mean(atr)
        if self._last_vol_reported < self.current_timestamp - self._volatility_interval:
            for market, vol in self._volatility.items():
                if not vol.is_nan():
                    self.logger().info(f"{market} volatility: {vol:.2%}")
            self._last_vol_reported = self.current_timestamp

    def notify_hb_app(self, msg: str):
        """
        Send a message to the hummingbot application
        """
        if self._hb_app_notification:
            super().notify_hb_app(msg)



import os
from typing import TYPE_CHECKING, List

import pandas as pd

from hummingbot.client.config.security import Security
from hummingbot.client.settings import DEFAULT_LOG_FILE_PATH
from hummingbot.core.utils.async_utils import safe_ensure_future
from hummingbot.model.trade_fill import TradeFill

if TYPE_CHECKING:
    from hummingbot.client.hummingbot_application import HummingbotApplication  # noqa: F401


class ExportCommand:
    def export(self,  # type: HummingbotApplication
               option):
        if option is None or option not in ("keys", "trades"):
            self.notify("Invalid export option.")
            return
        elif option == "keys":
            safe_ensure_future(self.export_keys())
        elif option == "trades":
            safe_ensure_future(self.export_trades())

    async def export_keys(self,  # type: HummingbotApplication
                          ):
        await Security.wait_til_decryption_done()
        if not Security.any_secure_configs():
            self.notify("There are no keys to export.")
            return
        self.placeholder_mode = True
        self.app.hide_input = True
        if await self.check_password():
            self.notify("\nWarning: Never disclose API keys or private keys. Anyone with your keys can steal any "
                        "assets held in your account.")
            self.notify("\nAPI keys:")
            for key, cm in Security.all_decrypted_values().items():
                for el in cm.traverse(secure=False):
                    if el.client_field_data is not None and el.client_field_data.is_secure:
                        self.notify(f"{el.attr}: {el.printable_value}")
        self.app.change_prompt(prompt=">>> ")
        self.app.hide_input = False
        self.placeholder_mode = False

    async def prompt_new_export_file_name(self,  # type: HummingbotApplication
                                          path):
        input = await self.app.prompt(prompt="Enter a new csv file name >>> ")
        if input is None or input == "":
            self.notify("Value is required.")
            return await self.prompt_new_export_file_name(path)
        if input == " ":
            return None
        if "." not in input:
            input = input + ".csv"
        file_path = os.path.join(path, input)
        if os.path.exists(file_path):
            self.notify(f"{input} file already exists, please enter a new name.")
            return await self.prompt_new_export_file_name(path)
        else:
            return input

    async def export_trades(self,  # type: HummingbotApplication
                            ):
        with self.trading_core.trade_fill_db.get_new_session() as session:
            trades: List[TradeFill] = self._get_trades_from_session(
                int(self.init_time * 1e3),
                session=session)
            if len(trades) == 0:
                self.notify("No past trades to export.")
                return
            self.placeholder_mode = True
            self.app.hide_input = True
            path = self.client_config_map.log_file_path
            if path is None:
                path = str(DEFAULT_LOG_FILE_PATH)
            file_name = await self.prompt_new_export_file_name(path)
            if file_name is None:
                return
            file_path = os.path.join(path, file_name)
            try:
                df: pd.DataFrame = TradeFill.to_pandas(trades)
                df.to_csv(file_path, header=True)
                self.notify(f"Successfully exported trades to {file_path}")
            except Exception as e:
                self.notify(f"Error exporting trades to {path}: {e}")
            self.app.change_prompt(prompt=">>> ")
            self.placeholder_mode = False
            self.app.hide_input = False



#!/usr/bin/env python

"""
Liquidity mining data types...
"""

from decimal import Decimal


class PriceSize:
    """
    Order price and order size.
    """

    def __init__(self, price: Decimal, size: Decimal):
        self.price: Decimal = price
        self.size: Decimal = size

    def __repr__(self):
        return f"[ p: {self.price} s: {self.size} ]"


class Proposal:
    """
    An order proposal for liquidity mining.
    market is the base quote pair like "ETH-USDT".
    buy is a buy order proposal.
    sell is a sell order proposal.
    """

    def __init__(self, market: str, buy: PriceSize, sell: PriceSize):
        self.market: str = market
        self.buy: PriceSize = buy
        self.sell: PriceSize = sell

    def __repr__(self):
        return f"{self.market} buy: {self.buy} sell: {self.sell}"

    def base(self):
        return self.market.split("-")[0]

    def quote(self):
        return self.market.split("-")[1]



"""
Tools for running asynchronous functions multiple times.
"""

import asyncio
import functools
import logging
from typing import Any, Dict, List, Optional, Type


class AllTriesFailedException(EnvironmentError):
    pass


def async_retry(retry_count: int = 2,
                exception_types: List[Type[Exception]] = [Exception],
                logger: logging.Logger = logging.getLogger("retry"),
                stats: Dict[str, int] = None,
                raise_exp: bool = True,
                retry_interval: float = 0.5
                ):
    """
    A decorator for async functions that will retry a function x times, where x is retry_count.

    :param retry_count: Number of retries
    :param exception_types: All exceptions trigger retry, but exceptions in the list also get logging
    :param logger: if raise_exp is false then log the last exception instead of raising it
    :param stats:
    :param raise_exp: raise an exception if all retries failed, otherwise log the last exception
    :param retry_interval: time to wait between retries
    """
    def decorator(fn):
        @functools.wraps(fn)
        async def retry(*args, _stats=stats, **kwargs):
            last_exception: Optional[Exception] = None
            for count in range(1, retry_count + 1):
                try:
                    additional_params: Dict[str, Any] = {}
                    fn_kwargs = {**additional_params, **kwargs}
                    return await fn(*args, **fn_kwargs)

                except tuple(exception_types) as exc:
                    last_exception = exc
                    logger.info(f"Exception raised for {last_exception}: {fn.__name__}. Retrying {count}/{retry_count} times.")
                    if _stats is not None and type(_stats) is dict:
                        metric_name: str = f"retry.{fn.__name__}.count"
                        if metric_name not in _stats:
                            _stats[metric_name] = 1
                        else:
                            _stats[metric_name] += 1
                except Exception as exc:
                    last_exception = exc
                    raise
                await asyncio.sleep(retry_interval)
            if raise_exp:
                raise AllTriesFailedException() from last_exception
            else:
                logger.info(f"Last exception raised for {repr(last_exception)}: {fn.__name__}. aborting.")
        return retry

    return decorator
